# -*- coding: utf-8 -*-

import os
from appdirs import user_config_dir, user_log_dir
from typing import Optional
from dataclasses import dataclass

DATA_DIR = os.path.normpath(os.path.expanduser("~/FRHEED"))
os.makedirs(DATA_DIR, exist_ok=True)

CONFIG_DIR = os.path.join(user_config_dir(), "FRHEED", "")
os.makedirs(CONFIG_DIR, exist_ok=True)

LOG_DIR = os.path.join(user_log_dir(), "FRHEED", "")
os.makedirs(LOG_DIR, exist_ok=True)

_LOCAL_DIR = os.path.dirname(__file__)
RESOURCE_DIR = os.path.normpath(os.path.join(_LOCAL_DIR, "resources", ""))
ICONS_DIR = os.path.normpath(os.path.join(RESOURCE_DIR, "icons", ""))
os.makedirs(ICONS_DIR, exist_ok=True)

WINDOW_ICON_PATH = os.path.join(ICONS_DIR, "FRHEED.ico")

# Colors from matplotlib tableau colors 
# Used matplotlib.colors.to_hex(...) for each color
# https://matplotlib.org/3.1.0/gallery/color/named_colors.html
COLOR_DICT = {
    "blue":     "#1f77b4",
    "orange":   "#ff7f0e",
    "green":    "#2ca02c",
    "red":      "#d62728",
    "purple":   "#9467bd",
    "brown":    "#8c564b",
    "pink":     "#e377c2",
    "gray":     "#7f7f7f",
    "olive":    "#bcbd22",
    "cyan":     "#17becf",
    }


def get_data_dir(
        user: Optional[str] = None, 
        experiment: Optional[str] = None
        ) -> str:
    """
    Create a data directory based on the given user and experiment name.

    Parameters
    ----------
    user : Optional[str], optional
        The name of the current user. If the default of None is used,
    no user subfolder will be created.
    experiment : Optional[str], optional
        The name of the current experiment. If the default of None is used,
    no experiment subfolder will be created.

    Returns
    -------
    str
        Path to the generated data directory.

    """
    
    data_dir = DATA_DIR
    
    if user is not None:
        data_dir = os.path.join(data_dir, user, "")
        
    if experiment is not None:
        data_dir = os.path.join(data_dir, experiment, "")
        
    os.makedirs(data_dir, exist_ok=True)
        
    return data_dir


if __name__ == "__main__":
    def test():
        print(f"Data directory: {DATA_DIR}")
        print(f"Config directory: {CONFIG_DIR}")
        print(f"Resource directory: {RESOURCE_DIR}")
        print(f"Icons directory: {ICONS_DIR}")
        print(f"Test directory: {get_data_dir('test', 'test')}")

    test()



# -*- coding: utf-8 -*-
"""
Connecting to FLIR cameras.
Adapted from simple_pyspin: https://github.com/klecknerlab/simple_pyspin
"""

from typing import Tuple, Union
import time
import numpy as np

from frheed.cameras import CameraError

# Make sure PySpin is installed
from frheed.cameras.flir.install_pyspin import install_pyspin
install_pyspin()
import PySpin


# Editable camera settings to show
_GUI_SETTINGS = {
    "AcquisitionFrameRate":         "Frame Rate",
    "AcquisitionFrameRateEnable":   "Enable Frame Rate",
    "BinningHorizontal":            "Horizontal Binning",
    "BinningHorizontalMode":        "Horizontal Binning Mode",
    "BinningVertical":              "Vertical Binning",
    "BinningVerticalMode":          "Vertical Binning Mode",
    "BlackLevel":                   "Black Level",
    "DeviceIndicatorMode":          "LED Behaviour",
    "DeviceUserID":                 "Camera Name",
    "ExposureAuto":                 "Auto-Exposure",
    "ExposureMode":                 "Exposure Mode",
    "ExposureTime":                 "Exposure Time",  # microseconds
    "GainAuto":                     "Auto-Gain",
    "Gain":                         "Gain",  # note: read-only if GainAuto is True
    "Gamma":                        "Gamma",
    "GammaEnable":                  "Enable Gamma",
    }

# Read-only camera information to show
_GUI_INFO = {
    "Width":                        "Frame Width",
    "SensorWidth":                  "Sensor Width",
    "WidthMax":                     "Maximum Frame Width",  # calculated after horizontal binning
    "Height":                       "Frame Height",
    "SensorHeight":                 "Sensor Height",
    "HeightMax":                    "Maximum Frame Height",  # calculated after vertical binning
    "DeviceVendorName":             "Vendor",
    "DeviceModelName":              "Model",
    "DeviceID":                     "Serial Number",
    "DeviceTLType":                 "Type",  # e.g. GigEVision
    "DeviceTemperature":            "Temperature",
    "DeviceUptime":                 "Uptime",
    "DeviceVersion":                "Firmware Version",
    "PixelDynamicRangeMin":         "Dynamic Range Minimum",  # e.g. 0
    "PixelDynamicRangeMax":         "Dynamic Range Maximum",  # e.g. 256
    "PixelFormat":                  "Pixel Format",
    "PixelSize":                    "Pixel Size",  # bits per pixel,
    "PowerSupplyCurrent":           "Current",
    "PowerSupplyVoltage":           "Voltage",
    "SerialPortBaudRate":           "Baud Rate",
    }

# Camera functions to show
_GUI_FUNCTIONS = {
    "DeviceReset":                  "Reboot",
    "FactoryReset":                 "Factory Reset",
    "SerialReceiveQueueClear":      "Clear Serial Port",
    }

_DEBUG = (__name__ == "__main__")

_SYSTEM = None


def list_cameras() -> PySpin.CameraList:
    """
    Return a list of Spinnaker cameras. Also initializes the PySpin
    'System', if needed. (See PySpin documentation for more info.)
    """

    global _SYSTEM

    if _SYSTEM is None:
        _SYSTEM = PySpin.System.GetInstance()

    return _SYSTEM.GetCameras()

def get_available_cameras() -> dict:
    """ Get available cameras as a dictionary of {source: name}. """
    cams = list_cameras()
    num_cams = cams.GetSize()
    available = {}
    
    for src in range(num_cams):
        try:
            with FlirCamera(src=src) as cam:
                if cam.initialized:
                    available[src] = str(cam)
                else:
                    print(f"FLIR camera {src} is not available")
        except CameraError:
            print("No FLIR cameras detected")
            
    return available


class FlirCamera:
    """
    A class used to encapsulate a PySpin camera.
    Attributes
    ----------
    cam : PySpin Camera
    running : bool
        True if acquiring images
    camera_attributes : dictionary
        Contains links to all of the camera nodes which are settable
        attributes.
    camera_methods : dictionary
        Contains links to all of the camera nodes which are executable
        functions.
    camera_node_types : dictionary
        Contains the type (as a string) of each camera node.
    lock : bool
        If True, attribute access is locked down; after the camera is
        initialized, attempts to set new attributes will raise an error.  This
        is to prevent setting misspelled attributes, which would otherwise
        silently fail to acheive their intended goal.
    intialized : bool
        If True, init() has been called.
        
    In addition, many more virtual attributes are created to allow access to
    the camera properties.  A list of available names can be found as the keys
    of 'camera_attributes' dictionary, and a documentation file for a specific
    camera can be genereated with the 'document' method.
    
    Methods
    -------
    init()
        Initializes the camera.  Automatically called if the camera is opened
        using a 'with' clause.
    close()
        Closes the camera and cleans up.  Automatically called if the camera
        is opening using a 'with' clause.
    start()
        Start recording images.
    stop()
        Stop recording images.
    get_image()
        Return an image using PySpin"s internal format.
    get_array()
        Return an image as a Numpy array.
    get_info(node)
        Return info about a camera node (an attribute or method).
    document()
        Create a Markdown documentation file with info about all camera
        attributes and methods.
        
    """

    _rw_modes = {
        PySpin.RO: "read only",
        PySpin.RW: "read/write",
        PySpin.WO: "write only",
        PySpin.NA: "not available"
        }

    _attr_types = {
        PySpin.intfIFloat:          PySpin.CFloatPtr,
        PySpin.intfIBoolean:        PySpin.CBooleanPtr,
        PySpin.intfIInteger:        PySpin.CIntegerPtr,
        PySpin.intfIEnumeration:    PySpin.CEnumerationPtr,
        PySpin.intfIString:         PySpin.CStringPtr,
        }

    _attr_type_names = {
        PySpin.intfIFloat:          "float",
        PySpin.intfIBoolean:        "bool",
        PySpin.intfIInteger:        "int",
        PySpin.intfIEnumeration:    "enum",
        PySpin.intfIString:         "string",
        PySpin.intfICommand:        "command",
        }

    def __init__(self, src: Union[int, str] = 0, lock: bool = False):
        """
        Parameters
        ----------
        src : int or str (default: 0)
            If an int, the index of the camera to acquire.  If a string,
            the serial number of the camera.
        lock : bool (default: False)
            If True, setting new attributes after initialization results in
            an error.
        """
        super().__setattr__("camera_attributes", {})
        super().__setattr__("camera_methods", {})
        super().__setattr__("camera_node_types", {})
        super().__setattr__("_initialized", False)
        super().__setattr__("lock", lock)

        cam_list = list_cameras()
        
        if _DEBUG: 
            print(f"Found {cam_list.GetSize()} FLIR camera(s)")

        self._src_type = type(src)            
        self._src = src
            
        if not cam_list.GetSize():
            raise CameraError("No FLIR cameras detected.")
            
        if isinstance(src, int):
            self._cam = cam_list.GetByIndex(src)
            
        elif isinstance(src, str):
            self._cam = cam_list.GetBySerial(src)
            
        cam_list.Clear()
        
        # Other attributes which may be accessed later
        self._running = False
        self._frame_times = []
        self._incomplete_image_count = 0

    def __getattr__(self, attr: str) -> object:
        # Add this in so @property decorator works as expected
        if attr in self.__dict__:
            return self.__dict__[attr]
        
        elif attr in self.camera_attributes:

            prop = self.camera_attributes[attr]
            if not PySpin.IsReadable(prop):
                raise CameraError(f"Camera property '{attr}' is not readable")

            if hasattr(prop, "GetValue"):
                return prop.GetValue()
            elif hasattr(prop, "ToString"):
                return prop.ToString()
            else:
                raise CameraError(f"Camera property '{attr}' is not readable")
                
        elif attr in self.camera_methods:
            return self.camera_methods[attr].Execute
        
        else:
            raise AttributeError(attr)

    def __setattr__(self, attr: str, val: object) -> None:
        if attr in self.camera_attributes:

            prop = self.camera_attributes[attr]
            if not PySpin.IsWritable(prop):
                raise CameraError(f"Property '{attr}' is not currently writable!")

            if hasattr(prop, "SetValue"):
                prop.SetValue(val)
            else:
                prop.FromString(val)

        elif attr in self.camera_methods:
            raise CameraError(f"Camera method '{attr}' is a function -- "
                              "you can't assign it a value!")
        else:
            if attr == "__class__":
                super().__setattr__(attr, val)
            elif attr not in self.__dict__ and self.lock and self.initialized:
                raise CameraError(f"Unknown property '{attr}'.")
            else:
                super().__setattr__(attr, val)
                
    def __enter__(self) -> "FlirCamera":
        self.init()
        return self
    
    def __exit__(self, type, value, traceback) -> None:
        self.close()
        
    def __del__(self) -> None:
        # Close the camera
        try:
            self.close()
            
        # If "with" clause is called and camera doesn't exist, AttributeError will happen
        except AttributeError:
            pass

    def __str__(self) -> str:
        model = getattr(self, "DeviceModelName", "Camera")
        return f"FLIR {model} (SN {self.DeviceSerialNumber})"

    @property
    def name(self) -> str:
        return f"FLIR{self._src}"
    
    @property
    def camera_type(self) -> str:
        return "FLIR"

    @property
    def cam(self) -> PySpin.PySpin.CameraPtr:
        return self._cam

    @property
    def initialized(self) -> bool:
        return self._initialized

    @property
    def running(self) -> bool:
        return self._running
    
    @property
    def incomplete_image_count(self) -> int:
        return self._incomplete_image_count

    @property
    def model(self) -> str:
        """ Camera model, including vendor and device model name """
        return f"{self.DeviceVendorName.strip()} {self.DeviceModelName.strip()}"

    @property
    def settings(self) -> dict:
        """ Get public, accessible camera settings as a dictionary """
        settings = {}
        for attr in sorted(self.camera_attributes.keys()):
            # Skip private attributes
            if "_" in attr:
                continue
            
            # Get attribute information
            info = self.get_info(attr)
            
            # Skip inaccessible attributes
            if not info.get("access", 0):
                continue
            
            # Add attribute to settings
            settings[attr] = info
            
        return settings
    
    @property
    def gui_settings(self) -> dict:
        """ User-editable attributes to show in a GUI """
        return _GUI_SETTINGS
    
    @property
    def gui_info(self) -> dict:
        """ Read-only attributes to show in a GUI """
        return _GUI_INFO
    
    @property
    def gui_functions(self) -> dict:
        """ User-accessible functions to show in a GUI """
        return _GUI_FUNCTIONS
    
    @property
    def real_fps(self) -> float:
        """ Get the real frames per second (Hz) """
        
        # When not enough frames have been captured
        if len(self._frame_times) <= 1:
            return 0.
        
        # When fewer than 60 frames have been captured in this acquisition
        elif len(self._frame_times) < 60:
            dt = (self._frame_times[-1] - self._frame_times[0])
            return len(self._frame_times) / max(dt, 1)
        
        # Return the average frame time of the last 60 frames
        else:
            return 60 / (self._frame_times[-1] - self._frame_times[-60])

    @property
    def width(self) -> int:
        if not self.initialized:
            self.init()
            width = self.Width
            self.close()
        
        else:
            width = self.Width
            
        return int(width)
    
    @property
    def height(self) -> int:
        if not self.initialized:
            self.init()
            height = self.Height
            self.close()
        
        else:
            height = self.Height
            
        return int(height)

    @property
    def shape(self) -> Tuple[int, int]:
        """ Get the camera array dimensions (Height x Width) """
        if not self.initialized:
            self.init()
            shape = (self.Height, self.Width)
            self.close()
            
        else:
            shape = (self.Height, self.Width)
        
        return shape

    def init(self) -> None:
        """
        Initializes the camera. Automatically called if the camera is opened
        using a 'with' clause.
        """

        self.cam.Init()

        for node in self.cam.GetNodeMap().GetNodes():
            pit = node.GetPrincipalInterfaceType()
            name = node.GetName()
            self.camera_node_types[name] = self._attr_type_names.get(pit, pit)
            if pit == PySpin.intfICommand:
                self.camera_methods[name] = PySpin.CCommandPtr(node)
            if pit in self._attr_types:
                self.camera_attributes[name] = self._attr_types[pit](node)

        self._initialized = True

    def start(self, continuous: bool = True) -> None:
        """
        Start capturing frames from the camera.

        Parameters
        ----------
        continuous : bool, optional
            Whether to capture frames continuously or one at a time. The default is True.

        Returns
        -------
        None
            DESCRIPTION.

        """
        
        # Set acquisition mode to "Continuous" (i.e. streaming)
        # This has to be done prior to initialization
        # TODO: Support other acquisition modes
        if continuous:
            self.AcquisitionMode = "Continuous"
            
        # Initialize the camera
        if not self.initialized:
            self.init()
            
        # Begin acquisition
        if not self.running:
            if not self.cam.IsStreaming():
                self.cam.BeginAcquisition()
            self._running = True

    def stop(self) -> None:
        """ Stop recording images. """
        
        if self.running:
            self.cam.EndAcquisition()
        self._frame_times = []
        self._incomplete_image_count = 0
        self._running = False
        
    def close(self) -> None:
        """
        Closes the camera and cleans up. Automatically called if the camera
        is opening using a 'with' clause.
        """

        # Stop the camera
        self.stop()
        try:
            del self.cam
        except AttributeError:
            pass
        
        # Reset attributes
        self.camera_attributes = {}
        self.camera_methods = {}
        self.camera_node_types = {}
        self._initialized = False
        # self.system.ReleaseInstance()

    def get_image(self, wait: bool = True) -> PySpin.ImagePtr:
        """
        Get an image from the camera.
        Parameters
        ----------
        wait : bool (default: True)
            If True, waits for the next image.  Otherwise throws an exception
            if there isn"t one ready.
        Returns
        -------
        img : PySpin Image
        """
        
        # Make sure the camera is running
        if not self.running:
            self.start()
        
        # Get the image pointer
        image_ptr = self.cam.GetNextImage(PySpin.EVENT_TIMEOUT_INFINITE if wait 
                                          else PySpin.EVENT_TIMEOUT_NONE)
        
        # Check if the image is incomplete
        if image_ptr.IsIncomplete():
            self._incomplete_image_count += 1
        
        # Release the image pointer to free memory (I think this reduces performance?)
        # image_ptr.Release()  # free memory in the camera buffer
        
        return image_ptr

    def get_array(
            self, 
            wait: bool = True, 
            get_chunk: bool = False, 
            complete_frames_only: bool = False
            ) -> Union[np.ndarray, Tuple[np.ndarray, PySpin.PySpin.ChunkData]]:
        
        """
        Get an image from the camera, and convert it to a numpy array.
        
        Parameters
        ----------
        wait : bool (default: True)
            If True, waits for the next image.  Otherwise throws an exception
            if there isn"t one ready.
        get_chunk : bool (default: False)
            If True, returns chunk data from image frame.
        complete_frames_only : bool (default: True)
            If True, only return complete frames.
            
        Returns
        -------
        img : numpy.ndarray
        chunk : PySpin.PySpin.ChunkData (only if get_chunk == True)
        """

        # Get image pointer
        img = self.get_image(wait=wait)
        
        # Ensure complete image is returned if option is chosen
        if complete_frames_only and img.IsIncomplete():
            return self.get_array(wait, get_chunk, complete_frames_only)
            
        # Store frame time for real FPS calculation
        self._frame_times.append(time.time())

        if get_chunk:
            return img.GetNDArray(), img.GetChunkData()
        else:
            self._frame_times.append(time.time())
            return img.GetNDArray()

    def get_info(self, name: str) -> dict:
        """
        Get information on a camera node (attribute or method).
        
        Parameters
        ----------
        name : string
            The name of the desired node
        Returns
        -------
        info : dict
            A dictionary of retrieved properties.  *Possible* keys include:
                - "access": read/write access of node.
                - "description": description of node.
                - "value": the current value.
                - "unit": the unit of the value (as a string).
                - "min" and "max": the min/max value.
        """
        info = {"name": name}

        if name in self.camera_attributes:
            node = self.camera_attributes[name]
        elif name in self.camera_methods:
            node = self.camera_methods[name]
        else:
            raise ValueError(f"'{name}' is not a camera method or attribute")

        info["type"] = self.camera_node_types[name]

        if hasattr(node, "GetAccessMode"):
            access = node.GetAccessMode()
            info["access"] = self._rw_modes.get(access, access)
            # print(info["access"])
            if isinstance(info["access"], str) and "read" in info["access"]:
                info["value"] = getattr(self, name)

        # print(info)
        if info.get("access") != 0:
            for attr in ("description", "unit", "min", "max"):
                fname = "Get" + attr.capitalize()
                f = getattr(node, fname, None)
                if f:
                    info[attr] = f()
            if hasattr(node, "GetEntries"):
                entries = []
                entry_desc = []
                has_desc = False
                for entry in node.GetEntries():
                    entries.append(entry.GetName().split("_")[-1])
                    entry_desc.append(entry.GetDescription().strip())
                    if entry_desc[-1]: has_desc = True
                info["entries"] = entries
                if has_desc:
                    info["entry_descriptions"] = entry_desc

        return info

    def document(self, verbose: bool = True) -> str:
        """
        Creates a MarkDown documentation string for the camera.

        Parameters
        ----------
        verbose : bool, optional
            Whether to show documentation details, including default access, 
            value, and range. The default is True.

        Returns
        -------
        str
            A string describing the camera attributes and methods.

        """
        
        # Get basic camera information, including model and version
        lines = [self.model]
        lines.append("=" * len(lines[-1]))
        lines.append("")
        lines.append("*Version: %s*" % getattr(self, "DeviceVersion", "?"))
        lines.append("")

        # Get camera attributes
        lines.append("Attributes")
        lines.append("-" * len(lines[-1]))
        lines.append("")
        
        for attr in sorted(self.camera_attributes.keys()):
            
            # Skip private attributes
            if "_" in attr:
                continue
            # print(attr)
            
            # Get attribute information
            info = self.get_info(attr)
            if not info.get("access", 0):
                continue
            lines.append(f"{attr}: {info.get('type', '?')}")
            
            # Skip details if not verbose (gives shorter markdown document)
            if not verbose:
                continue
            
            # Get attribute description
            lines.append("  " + info.get("description", "(no description provided)"))
            
            # Get default access
            lines.append("  - default access: %s" % info.get("access", "?"))
            if "value" in info:
                lines.append("  - default value: %s" % repr(info["value"]))
            if "unit" in info and info["unit"].strip():
                lines.append("  - unit: %s" % info["unit"])
            if "min" in info and "max" in info:
                lines.append("  - default range: %s - %s" % (info["min"], info["max"]))
            if "entries" in info:
                if "entry_descriptions" in info:
                    lines.append("  - possible values:")
                    for e, ed in zip(info["entries"], info["entry_descriptions"]):
                        if ed:
                            lines.append("    - '%s': %s" % (e, ed))
                        else:
                            lines.append("    - '%s'" % e)
                else:
                    lines.append("  - possible values: %s" % 
                                 (", ".join("'%s'" % e for e in info["entries"])))

            lines.append("")
        
        # Get camera methods
        lines.append("")
        lines.append("Methods")
        lines.append("-" * len(lines[-1]))
        lines.append("")
        lines.append(("**Note: the camera recording should be started/stopped"
                      " using the 'start' and 'stop' methods, not any of the functions"
                      " below (see simple_pyspin documentation).**"))
        lines.append("")

        for attr in sorted(self.camera_methods.keys()):
            
            # Skip private methods
            if "_" in attr:
                continue
            # print(attr)
            
            # Get method information
            info = self.get_info(attr)
            lines.append("%s():  " % (attr))
            
            # Get method description
            lines.append("  " + info.get("description", "(no description provided)"))
            
            # Skip details if not verbose (gives shorter markdown document)
            if not verbose:
                lines.append("")
                continue
            
            # Get method default access
            lines.append("  - default access: %s" % info.get("access", "?"))
            lines.append("")

        return "\n".join(lines)
    

if __name__ == "__main__":
    def test():
        with FlirCamera() as cam:
            print(cam.document(verbose=False))
            print(cam)
            cam.start()
            while True:
                try:
                    global image
                    image = cam.get_array()
                    # print(cam.incomplete_image_count, cam.real_fps)
                    print(cam.DeviceTemperature)
                    
                except KeyboardInterrupt:
                    break
    
    # test()
    
    print(list_cameras())
    


# -*- coding: utf-8 -*-
"""
General utility functions for FRHEED.
"""

import sys
import os
from typing import Union, Optional, Dict, Tuple
import logging
import inspect
from pathlib import Path
import subprocess

import numpy as np
from PyQt5.QtWidgets import QWidget, QApplication
from PyQt5.QtGui import QColor, QPen, QIcon

from frheed import settings
from frheed.constants import LOG_DIR


_DEBUG = (__name__ == "__main__")


def get_logger(name: Optional[str] = None) -> logging.Logger:
    """ Get a logger for FRHEED errors and messages. """
    # Use name if provided, otherwise get name of module that called this function
    name = name or Path(inspect.getmodule(inspect.stack()[1][0]).__file__).stem
    
    # Generate log path
    filepath = os.path.join(LOG_DIR, f"{name}.log")
    
    # Create logger
    logger = logging.getLogger(name=name)
    
    # Make sure handlers haven't been added already
    # https://stackoverflow.com/a/59448231/10342097
    if logger.handlers:
        logger.info(f"{name}.log is already running")
        return logger
    
    # Set level to info
    logger.setLevel(logging.INFO)
    
    # Filter that logs only INFO and WARNING level messages
    class ConsoleFilter:
        def __init__(self, level: int):
            self.__level = level
            
        def filter(self, record) -> bool:
            return record.levelno in [logging.INFO, logging.WARNING]
    
    # Create file handler
    file_handler = logging.FileHandler(filepath, mode="a")
    file_format = "%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s] %(message)s"
    file_formatter = logging.Formatter(file_format)
    file_handler.setFormatter(file_formatter)
    logger.addHandler(file_handler)
    
    # Create console handler
    console_handler = logging.StreamHandler()
    console_format = "%(message)s"
    console_formatter = logging.Formatter(console_format)
    console_handler.setFormatter(console_formatter)
    console_handler.addFilter(ConsoleFilter(logging.INFO))
    logger.addHandler(console_handler)
    
    # Notify that logfile was started
    logger.info(f"Started {name}.log")
    
    return logger


logger = get_logger("utils")


def get_platform_bitsize() -> int:
    """
    Get the Windows platform bit size (32 or 64).

    Returns
    -------
    int
        Windows architecture bit size (either 32 or 64).

    """
    import struct
    
    bitsize = struct.calcsize("P") * 8
    
    if _DEBUG:
        print(f"Platform bitsize: {bitsize}")
        
    return bitsize
    

def fix_pyqt() -> None:
    """ Fixes system excepthook for QApplication instances. """
    # https://stackoverflow.com/a/47275100/3620725
    
    sys._excepthook = sys.excepthook
    
    def pyqt_except_hook(exctype, value, traceback):
        print(exctype, value, traceback)
        sys._excepthook(exctype, value, traceback)
        sys.exit(1)
        
    sys.excepthook = pyqt_except_hook
    
    
def fix_ipython() -> None:
    """ Prevents PyQt5 from becoming unresponsive in IPython outside main loop. """
    # Uses pandasgui implementation from pandasgui.utility
    from IPython import get_ipython
    ipython = get_ipython()
    if ipython is not None:
        return ipython.magic("gui qt5")
    

def fit_screen(widget: QWidget, scale: float = 0.5) -> None:
    """ Fit a widget in the center of the main screen """
    
    from PyQt5.QtWidgets import QDesktopWidget
    
    # Get main screen geometry
    desktop = QDesktopWidget()
    screen = desktop.availableGeometry()
    tot_w, tot_h = screen.width(), screen.height()
    
    # Resize widget to 75% of available space
    w, h = int(tot_w * scale), int(tot_h * scale)
    widget.resize(w, h)
    
    # Move to center of screen
    dx, dy = int((tot_w - w) / 2), int((tot_h - h) / 2)
    widget.move(dx, dy)
    
    if _DEBUG:
        print(f"Resized widget to {widget.width()}, {widget.height()} "
              f"and moved it to {dx}, {dy}")
    
    # Show the widget
    # widget.show()


def get_icon(name: str) -> QIcon:
    """ Get the icon with the given name from the icons directory as a QIcon. """
    from frheed.constants import ICONS_DIR
    return QIcon(os.path.join(ICONS_DIR, f"{name}.ico"))


def test_widget(
        widget_class: type, 
        block: bool = False,
        **kwargs
        ) -> Tuple[QWidget, QApplication]:
    """
    Create a widget from the provided class using the *args and **kwargs,
    and start a blocking application event loop if block = True.

    Parameters
    ----------
    widget_class : type
        Class of the widget to be created (NOT the widget itself).
    block : bool, optional
        Whether or not to start the blocking event loop automatically. 
        The default is False.
    *args
        Positional arguments to be passed to the widget_class.
    **kwargs
        Keyword arguments to be passed to the widget_class.

    Returns
    -------
    Tuple[QWidget, QApplication]
        A tuple containing the widget and QApplication instance.

    """
    
    from PyQt5.QtCore import Qt
    from PyQt5.QtWidgets import QLabel
    
    # Fix PyQt
    fix_pyqt()
    
    # Fix IPython
    fix_ipython()
    
    # Get QApplication instance
    app = QApplication.instance() or QApplication(["FRHEED"])
    app.setStyle(settings.APP_STYLE)
    
    # Create widget
    # NOTE: This MUST be done after creating the application instance
    # otherwise the Spyder kernel will crash without a traceback
    if widget_class is not None:
        widget = widget_class(**kwargs)
    else:
        widget = QLabel()
        widget.setText("Demo Widget")
        widget.setAlignment(Qt.AlignCenter)
    
    # Set window icon
    widget.setWindowIcon(get_icon("FRHEED"))
    
    # Set window title to the widget class name
    widget.setWindowTitle(widget.__class__.__name__)
    
    # This should be called by default, but just make sure
    app.lastWindowClosed.connect(app.quit)
    
    # Fit to screen
    fit_screen(widget)
    
    # Show the widget
    widget.show()
    
    # NOTE: _exec() starts a blocking loop; any code after will not run
    sys.exit(app.exec_()) if block else None
    
    return (widget, app)


def unit_string(
        value: Union[float, int], 
        unit: str, 
        sep: Optional[str] = None,
        precision: Optional[int] = 2
        ) -> str:
    """
    Format a unit string depending on the order of magnitude.
    For example: 3_000_000 µm -> 3 m if base_unit = "m".

    Parameters
    ----------
    value : Union[float, int]
        The value to format as a string with appropriate units.
    unit : str
        Unit, such as "m" for meters or "µm" for microns.
    sep : Optional[str], optional
        Separator between the value and the suffix. If the default of None
        is used, the separator will be a space (" ").
    precision : Optional[int], optional
        Floating point precision to format as. If the default of None is 
        used, the inferred format "{...:,g}" will be used.

    Returns
    -------
    str
        A formatted string displaying the value with appropriate units.

    """
    
    from math import floor, log10
    
    # Certain units should have particular specifiers
    no_space_units = ["%"]
    sep = sep or " "
    if unit and any(v in unit for v in no_space_units):
        sep = ""
    
    # If unit is specific one, return early
    if unit in ["dB", "Hz", "%"]:
        if precision:
            return f"{value:.{precision}f}{sep}{unit}"
        else:
            return f"{value:,g}{sep}{unit}"
    
    # If no unit is given, just format using :,g
    if not unit:
        unit_str = f"{value:,g}" if precision is None else f"{value:.{precision}f}"
        return unit_str
        
    # Get prefixes for magnitudes, e.g. 10^-9 -> "n" for "nano"
    prefixes = {
        -9:     "n",  # nano
        -6:     "µ",  # micro
        -3:     "m",  # milli
        -2:     "c",  # centi
        -1:     "d",  # deci
        0:      "",   # base units
        3:      "k",  # kilo
        6:      "M",  # mega
        9:      "G",  # giga
        }
    
    # For certain units, like seconds, only certain prefixes should be used
    if unit and unit[-1] == "s":
        prefixes = {mag: s for mag, s in prefixes.items() if mag % 3 == 0}
    
    # Get magnitudes from prefices
    magnitudes = {prefix: mag for mag, prefix in prefixes.items()}
    
    # Add entry for things like "u" instead of "µ"
    magnitudes["u"] = -6
    
    # Make sure value is > 0 so log is valid
    orig_value = value
    if value < 0:
        value = abs(value)
        sign = "-"
    else:
        sign = ""
    
    # Make sure unit is only a single character in case something like µm is given
    # and the first character is in magnitudes
    unit_type = unit[-1] if unit[0] in magnitudes else unit
    
    # Scale value if unit already has a magnitude, e.g. µs
    if len(unit) > 1:
        value *= (10 ** magnitudes.get(unit[0], 0))

    # Get order of magnitude
    if value != 0:
        magnitude = floor(log10(value))
    else:
        magnitude = 0
    
    # If magnitude is not in prefixes, find the closest prefix
    if magnitude not in prefixes:
        magnitude = sorted(prefixes.keys(), key=lambda p: abs(magnitude - p))[0]
        
    # Scale value by determined magnitude
    scaled_value = value / (10 ** magnitude)
    
    # Get unit prefix
    prefix = prefixes[magnitude]
    
    # Generate unit string
    if precision is None:
        unit_str = f"{sign}{scaled_value:,g}{sep}{prefix}{unit_type}"
    else:
        unit_str = f"{sign}{scaled_value:.{precision}f}{sep}{prefix}{unit_type}"
    
    if _DEBUG:
        print(f"Converted {orig_value:,g} {unit} -> {unit_str}")
        
    return unit_str


def save_settings(
        settings: Dict[str, Dict[str, Union[bool, str, float, int]]], 
        name: str
        ) -> None:
    """
    Save a dictionary of settings to a .json file.

    Parameters
    ----------
    settings : Dict[str, Dict[str, Union[bool, str, float, int]]]
        A dictionary of the settings.
    name : str
        The name of the settings group to be saved.

    """
    import json
    from frheed.constants import CONFIG_DIR
    
    # Create dictionary with each setting represented by as dictionary
    # containing the value and type of that value so it can be converted back
    config = {}
    for group_name, setting_dict in settings.items():
        config[group_name] = {}
        for setting, value in setting_dict.items():
            config[group_name][setting] = {
                "value": value, 
                "type": type(value).__name__
                }
    
    # Get filepath
    path = os.path.join(CONFIG_DIR, f"{name}_settings.json")
    
    # Save the configuration file
    with open(path, "w") as f:
        json.dump(config, f, indent="\t")
        
    if _DEBUG:
        print(f"Successfully saved {name} settings:")
        for name, setting_dict in config.items():
            for setting, d in setting_dict.items():
                print(f"    {setting}:")
                for k, v in d.items():
                    print(f"        {k}: {v}")


def load_settings(name: str) -> Dict[str, Dict[str, Union[bool, str, float, int]]]:
    """
    Load a .json file into a dictionary of settings with proper types.

    Parameters
    ----------
    name : str
        The name of the settings group to be loaded.

    Returns
    -------
    dict
        Dictionary containing the settings with values converted 
    to their proper types.

    """
    import json
    from ast import literal_eval
    from frheed.constants import CONFIG_DIR
    
    # Get filepath
    path = os.path.join(CONFIG_DIR, f"{name}_settings.json")
    
    # This will raise an OSError if the file doesn't exist or is inaccessible
    with open(path, "r") as f:
        json_dict = json.load(f)

    # Convert to proper types
    config = {}
    for group_name, setting_dict in json_dict.items():
        
        config[group_name] = {}
        
        for setting, info in setting_dict.items():
            string_value = info["value"]
            type_ = info["type"]

            # json will natively convert some objects
            if not isinstance(string_value, str):
                config[group_name][setting] = string_value
                continue

            # Convert booleans
            if type_ == "bool":
                value = literal_eval(string_value)
                
            # Keep strings as strings
            elif type_ in ["string", "str"]:
                value = string_value
                
            # Convert integers and floats
            elif type_ in ["float", "int"]:
                value = literal_eval(string_value)
                
            # Try to convert if some other type is specified
            else:
                try:
                    value = literal_eval(string_value)
                except ValueError:
                    print(f"Unable to convert {string_value} to {type_}")
                    value = string_value
                    
            # Store updated type
            config[group_name][setting] = value
            
    if _DEBUG:
        print(f"Successfully loaded {name} settings:")
        for name, setting_dict in config.items():
            print(f"    {name}:")
            for setting, value in setting_dict.items():
                print(f"        {setting}: {value}")
        
    return config


def sample_array(
            w: int = 2048, 
            h: int = 1536, 
            channels: int = 3, 
            dtype: str = "uint8"
            ) -> np.ndarray:
    """
    Generate a sample numpy array with the given dimensions and dtype.

    Parameters
    ----------
    w : int, optional
        Width of the array. The default is 2048.
    h : int, optional
        Height of the array. The default is 1536.
    channels : int, optional
        Number of channels in the array. The default is 3.
    dtype : str, optional
        Data type of the array. The default is "uint8".

    Returns
    -------
    numpy.ndarray
        The resulting array with shape (h, w, channels) of type "dtype".

    """
    
    # Get the image shape
    shape = (h, w, channels) if channels > 1 else (h, w)
    
    # Create the array
    arr = (np.random.rand(*shape) * 255).astype(dtype)
    
    if _DEBUG:
        print(f"Input shape ({w}, {h}, {channels}) -> array {arr.shape}\n{arr}")
    return arr

def get_qcolor(color: Union[str, tuple, QColor]) -> QColor:
    """ Create a QColor. See https://doc.qt.io/qt-5/qcolor.html """
    
    if isinstance(color, str):
        return QColor(color)
    
    elif isinstance(color, tuple):
        return QColor(*color)
    
    elif isinstance(color, QColor):
        return color
    
    else:
        raise TypeError(f"Unsupported color type {type(color)}")


def get_qpen(color: Union[str, tuple, QColor], cosmetic: bool = True) -> QColor:
    """ Create a QPen with the specified color """
    pen = QPen(get_qcolor(color))
    pen.setCosmetic(cosmetic)
    return pen


def snip_lists(*lists) -> list:
    min_len = min(map(len, lists))
    return [L[:min_len] for L in lists]
    

def get_locals(frame) -> dict:
    return dict(frame.f_back.f_locals.items())


def install_whl(filepath: str) -> int:
    """ Install a module using pip, either from a PyPi library or local file. """
    # Make sure .whl filepath and python.exe filepath are single-escaped
    python_path = sys.executable.replace("\\", "/")
    filepath = filepath.replace("\\", "/")
    
    # Use subprocess to execute the commands
    args = [python_path, "-m", "pip", "install", filepath]
    exit_code = subprocess.check_call(args)
    return exit_code


def gen_reqs() -> str:
    """ Create the requirements.txt file for FRHEED. """
    python = sys.executable.replace("\\", "/")
    subprocess.check_call([python, "-m", "pip", "freeze", ">", "requirements.txt"])
    requirements = Path("requirements.txt").read_text()
    logger.info(f"Generated requirements:\n{requirements}")


if __name__ == "__main__":
    def test():
        fix_pyqt()
        get_platform_bitsize(); print("")
        unit_string(1e6, "µs"); print("")
        
        test_settings = {
            "test_1": {
                "boolean": True,
                "string": "foo",
                "float": 3.33,
                "integer": 1, 
                }
            }
        save_settings(test_settings, "test"); print("")
        load_settings("test"); print("")
        
        sample_array(); print("")
        
        widget, app = test_widget(None)
        sys.exit(app.exec_())

    test()
    



# -*- coding: utf-8 -*-
"""
Functions for computing values from plots.
"""

from typing import Union, Optional

import numpy as np
from scipy.signal import find_peaks

from frheed.utils import snip_lists


# Ignore numpy warnings
np.seterr("ignore")


def calc_fft(x: list, y: list) -> tuple:
    """
    Calculate the FFT of a 1D series.

    Parameters
    ----------
    x : list
        X values.
    y : list
        Y values.

    Returns
    -------
    tuple
        A tuple containing (frequency, PSD) lists. PSD is Power Spectral Density.

    """
    # Make sure data is equal lengths
    # Note: this is probably unnecessary if pulling 
    # data directly from another plot
    x, y = snip_lists(x, y)
    
    # Return if x or y is invalid
    def invalid_data(data):
        return len(data) == 0 or np.nan in data
    if any(invalid_data(d) for d in (x, y)):
        return None, None
    
    # Create evenly-spaced list of sample points 
    numsamples = len(x)
    samplespacing = (x[-1]-x[0])/numsamples
    
    # Generate array of frequencies
    try:
        freq = np.fft.rfftfreq(numsamples, d=samplespacing)
    except:
        return None, None
    
    # Convert y to float32 to avoid type conflict error in following operation
    y = np.array(y, dtype=np.float32)
    
    # Remove DC signal from the y-data
    y -= np.mean(y)
    
    # Apply Hanning filter to smooth edge discontinuities
    window = np.hanning(numsamples+1)[:-1]
    if len(y) != len(window):
        return None, None
    hann = y*window

    # Calculate real FFT
    fftdata = np.fft.rfft(hann)
    
    # Normalize FFT data & catch warnings (RuntimeError) as exceptions
    import warnings
    with warnings.catch_warnings():
        warnings.filterwarnings("error")
        try:
            psd = abs(fftdata)**2/(np.abs(hann)**2).sum()
            psd = (psd*2)**0.5
            
        except Warning:
            return None, None
    
    # Sometimes the arrays can become different lengths and throw errors
    freq, psd = snip_lists(freq, psd)
    
    return (freq, psd)

def apply_cutoffs(
        x: list, 
        y: list, 
        minval: Optional[float] = None, 
        maxval: Optional[float] = None
        ) -> tuple:
    """ Return data that falls between a certain range. Pass None to use min or max. """
    
    # Return if x or y is empty
    if len(x) + len(y) == 0:
        return (x, y)
    x, y = map(np.array, (x, y))
    minval = minval or min(x)
    maxval = maxval or max(x)
    orig_x = x.copy()
    mask = (orig_x >= minval) & (orig_x <= maxval)
    return (x[mask], y[mask])

def detect_peaks(
        x: Union[list, tuple, np.ndarray],
        y: Union[list, tuple, np.ndarray],
        min_freq: Optional[float] = 0.0
        ) -> list:
    
    # Catch numpy RuntimeWarning as exceptions
    import warnings
    with warnings.catch_warnings():
        warnings.filterwarnings("error")
        
        try:
            # Filter to minimum frequency
            x, y = apply_cutoffs(x=x, y=y, minval=min_freq)
            
            # Height
            height = max(np.median(y) + 3*np.std(y), 1.5)
            
            # Threshold (vertical distance to neighbors)
            threshold = None
            
            # Distance between peaks (# of indices)
            distance = 50
            
            # Prominence
            prominence = None
            
            # Find peaks
            peak_indices, props = find_peaks(
                                        y, 
                                        height = height,
                                        threshold = threshold,
                                        distance = distance,
                                        prominence = prominence
                                        )
            
            # Get corresponding x-coordinates
            return [x[idx] for idx in peak_indices]
        
        except Warning:
            return


if __name__ == "__main__":
    pass



# Version change guidelines:
#   - Increment the MAJOR version when making incompatible API changes
#   - Increment the MINOR version when adding backwards-compatible functionality
#   - Increment the PATCH version when making backwards-compatible bug fixes
VERSION = (0, 0, 1)  # (MAJOR, MINOR, PATCH)
__version__ = ".".join(map(str, VERSION))


# -*- coding: utf-8 -*-
"""
Material properties for use in computations.
"""

from dataclasses import dataclass
from enum import Enum


# Reference (Table 30.1):
# Adachi, S. (2017). III-V Ternary and Quaternary Compounds. 
# Springer Handbook of Electronic and Photonic Materials, 1–1. 
# doi:10.1007/978-3-319-48933-9_30 

BINARY_LATTICE_PARAMETERS = {  # at 300 K
    "AlN": {
        "Zinc blende": {
            "a": 4.38,  # in Angstroms
            },
        "Wurtzite": {
            "a": 3.112,
            "c": 4.982,
            }
        },
    "AlP": {
        "Zinc blende": {
            "a": 5.4635,
            },
        },
    "AlAs": {
        "Zinc blende": {
            "a": 5.66139,
            },
        },
    "AlSb": {
        "Zinc blende": {
            "a": 6.1355,
            },
        },
    "α-GaN": {  # alpha (α) is Alt+224
        "Wurtzite": {
            "a": 3.1896,
            "c": 5.1855,
            }
        },
    "ß-GaN": {  # beta (ß) is Alt+225
        "Zinc blende": {
            "a": 4.52,
            },
        },
    "GaP": {
        "Zinc blende": {
            "a": 5.4508,
            },
        },
    "GaAs": {
        "Zinc blende": {
            "a": 5.65330,
            },
        },
    "GaSb": {
        "Zinc blende": {
            "a": 6.09593,
            },
        },
    "InN": {
        "Zinc blende": {
            "a": 4.986,
            },
        "Wurtzite": {
            "a": 3.548,
            "c": 5.760,
            }
        },
    "InP": {
        "Zinc blende": {
            "a": 5.8690,
            },
        },
    "InAs": {
        "Zinc blende": {
            "a": 6.0583,
            },
        },
    "InSb": {
        "Zinc blende": {
            "a": 6.47937,
            },
        },
    }


def compute_ternary_parameters():
    pass # TODO
    

class CrystalStructure(Enum):
    WURTZITE = "Wurtzite"
    ZINC_BLENDE = "Zinc blende"


@dataclass
class Material:
    compound: str
    structure: CrystalStructure
    
    

if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-

import sys

from PyQt5.QtWidgets import QMainWindow, QApplication

from frheed.widgets.rheed_widgets import RHEEDWidget
from frheed import utils
# from frheed import settings


logger = utils.get_logger()
windows = []

# Fix IPython and PyQt
utils.fix_ipython()
utils.fix_pyqt()


class FRHEED(QMainWindow):
    def __init__(self):
        # Get application BEFORE initializing
        self.app = QApplication.instance() or QApplication(sys.argv)
        
        # Initialize window
        super().__init__(parent=None)
        
        # Store reference so the window doesn't get garbage-collected
        windows.append(self)
        
        # Create the main widget
        self.rheed_widget = RHEEDWidget()
        self.setCentralWidget(self.rheed_widget)
        
        # Connect signals
        self.app.lastWindowClosed.connect(self.app.quit)
        
        # Set window properties
        self.setWindowTitle("FRHEED")
        self.setWindowIcon(utils.get_icon("FRHEED"))
        
        # Show the window and center in the screen
        self.show()
        utils.fit_screen(self)
        
        # Start blocking event loop that ends when app is closed
        sys.exit(self.app.exec_())
        

def show() -> FRHEED:
    logger.info("Opening FRHEED...")
    return FRHEED()


if __name__ == "__main__":
    gui = show()
    


# -*- coding: utf-8 -*-
"""
Widgets for selecting things, including the source camera to use.
"""

from typing import Optional, Union, List
from dataclasses import dataclass
from enum import Enum

from PyQt5.QtWidgets import (
    QWidget,
    QPushButton,
    QGridLayout,
    
    )
from PyQt5.QtCore import (
    Qt,
    pyqtSignal,
    
    )

from frheed.cameras.flir import FlirCamera, get_available_cameras as get_flir_cams
from frheed.cameras.usb import UsbCamera, get_available_cameras as get_usb_cams
from frheed.cameras import CameraError
from frheed.utils import get_icon


class CameraClasses(Enum):
    flir = FlirCamera
    usb = UsbCamera


@dataclass
class CameraObject:
    cam_class: object
    src: Union[str, int]
    name: Optional[str] = None
    
    def get_camera(self) -> Union[FlirCamera, UsbCamera]:
        return self.cam_class(src=self.src)


class CameraSelection(QWidget):
    camera_classes = (
        FlirCamera,
        UsbCamera
        )
    camera_selected = pyqtSignal()
    
    def __init__(self):
        super().__init__(None)
        
        # NOTE: No parent is provided so the window can be minimized to the taskbar
        # TODO: Apply global stylesheet
        
        # Attributes to be assigned later
        self._cam: Optional[FlirCamera, UsbCamera] = None
        
        # Check for available cameras
        cams = self.available_cameras()
        
        # Set window properties
        self.setWindowFlags(Qt.WindowStaysOnTopHint | Qt.Window)
        self.setWindowTitle("Select Camera")
        self.setWindowIcon(get_icon("FRHEED"))
        
        # Set size
        self.setMinimumWidth(300)
        
        # Create layout
        self.layout = QGridLayout()
        self.setLayout(self.layout)
        
        # If there are no cameras, no buttons need to be added
        if not cams:
            btn = QPushButton("No cameras found")
            btn.setEnabled(False)
            self.layout.addWidget(btn, 0, 0)
        
        # Create buttons for each camera
        for i, cam in enumerate(cams):
            # Create the button
            btn = QPushButton(cam.name)
            
            # Connect signal
            # Doing it this way is necessary because otherwise all lambda
            # functions will initialize the same camera
            def make_lambda(cam_obj: CameraObject):
                return lambda: self.select_camera(cam_obj)
            btn.clicked.connect(make_lambda(cam))
        
            # Add button to layout
            self.layout.addWidget(btn, i, 0)
            
        # Show the widget
        self.setVisible(True)
        
    def available_cameras(self) -> List[CameraObject]:
        # Check each camera class for availability
        usb_cams = [CameraObject(UsbCamera, src, name) 
                    for src, name in get_usb_cams().items()]
        flir_cams = [CameraObject(FlirCamera, src, name)
                     for src, name in get_flir_cams().items()]
        return usb_cams + flir_cams
    
    def select_camera(self, cam: CameraObject) -> object:
        """ Get the selected camera class object. """
        # Deselect existing camera
        try:
            self._cam.close()
        except:
            pass
        
        # Initialize camera
        self._cam = cam.get_camera()
        print(f"Connected to {cam.name}")
        
        # Emit camera_selected signal
        self.camera_selected.emit()
        
        # Hide the selection widget
        self.setVisible(False)
        
        return self._cam
    

if __name__ == "__main__":
    cam_select = CameraSelection()



# -*- coding: utf-8 -*-
"""
Widgets for plotting data in PyQt.
"""

from typing import Union, Optional

from PyQt5.QtWidgets import (
    QWidget,
    QGridLayout,
    QComboBox,
    QLabel,
    QPushButton,
    QSplitter,
    QMenuBar,
    QMenu,
    QAction,
    QWidgetAction,
    QDoubleSpinBox,
    QSizePolicy,
    QCheckBox,
    QGraphicsPixmapItem,
    
    )
from PyQt5.QtGui import (
    QColor,
    QFont,
    QPalette,
    QPen,
    QPixmap,
    
    )
from PyQt5.QtCore import (
    Qt,
    pyqtSlot,
    pyqtSignal,
    
    )
import pyqtgraph as pg  # import *after* PyQt5
import numpy as np

from frheed.widgets.common_widgets import HSpacer, VisibleSplitter
from frheed.widgets.camera_widget import DEFAULT_CMAP
from frheed.utils import get_qcolor, get_qpen, get_logger
from frheed.calcs import calc_fft, detect_peaks, apply_cutoffs
from frheed.image_processing import ndarray_to_qpixmap, apply_cmap


# https://pyqtgraph.readthedocs.io/en/latest/_modules/pyqtgraph.html?highlight=setConfigOption
_PG_CFG = {
    "leftButtonPan":        True,
    "foreground":           (0, 0, 0, 255),
    "background":           (0, 0, 0, 0),  # makes the background transparent
    "antialias":            False,
    "editorCommand":        None,
    "useWeave":             False,
    "weaveDebug":           True,
    "exitCleanup":          True,
    "enableExperimental":   False,
    "crashWarning":         True,
    "imageAxisOrder":       "col-major",
    "useOpenGL":            False,
    }
_PG_PLOT_STYLE = {
    "tickTextWidth":        30,
    "tickTextHeight":       18,
    "autoExpandTextSpace":  True,
    "tickFont":             None,
    "stopAxisAtTick":       (False, False),  ## Check this one
    "showValues":           True,
    "tickLength":           5,
    "maxTickLevel":         2,
    "maxTextLevel":         2,
    }
_PG_AXES = ("left", "bottom", "right", "top")
_AXIS_COLOR = QColor("black")
_DEFAULT_SIZE = (800, 600)
_MIN_FFT_PEAK_POS = 0.5
_CURVE_MENU_TITLE = "View Lines"
_ITALIC_COORDS = True
logger = get_logger()


def init_pyqtgraph(use_opengl: bool = False) -> None:
    """ Set up the pyqtgraph configuration options """
    for k, v in _PG_CFG.items():
        try:
            pg.setConfigOption(k, v)
        except Exception as ex:
            logger.exception(str(ex))
    

class PlotWidget(QWidget):
    """ The base plot widget for embedding in PyQt5 """
    curve_added = pyqtSignal(str)
    curve_removed = pyqtSignal(str)
    curve_toggled = pyqtSignal(str, bool)
    data_changed = pyqtSignal(str)
    
    def __init__(
            self, 
            parent: QWidget = None, 
            popup: bool = False, 
            name: Optional[str] = None,
            title: Optional[str] = None,
            show_menubar: bool = True,
            ):
        
        super().__init__(parent)
        self._parent = parent
        self.name = name
        self.title = title
        self.show_menubar = show_menubar
        
        # Settings
        init_pyqtgraph()
        
        # Create layout
        self.layout = QGridLayout()
        self.layout.setContentsMargins(8, 8, 8, 8)
        self.layout.setSpacing(4)
        self.setLayout(self.layout)
        
        # Create plot widget
        self.plot_widget = pg.PlotWidget(self, title=title, background=_PG_CFG["background"])
        for ax in [self.plot_widget.getAxis(a) for a in _PG_AXES]:
            ax.setPen(_AXIS_COLOR)
            ax.setStyle(**{**_PG_PLOT_STYLE, **{"tickFont": self.font()}})
        
        # Create menubar with transparent background
        self.menubar = QMenuBar(self)
        self.menubar.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
        self.setStyleSheet(self.styleSheet() + "QMenuBar { background-color: transparent; }")
        
        # Create menu for showing/hiding curves
        self.curve_menu = self.menubar.addMenu(_CURVE_MENU_TITLE)
        # self.curve_menu.setStyleSheet(self.curve_menu.styleSheet() + "QMenu { font-weight: bold; }")
        
        # Create cursor label
        self.cursor_label = QLabel()
        self.cursor_label.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        self.cursor_label.setMinimumWidth(64)
        if _ITALIC_COORDS:
            self.cursor_label.setStyleSheet(self.cursor_label.styleSheet() + "font-style: italic")
        
        # Attributes to be assigned/updated later
        self.plot_items = {}
        
        # Add widgets (leave space in the middle for other widgets)
        self.layout.addWidget(self.menubar, 0, 0, 1, 1, Qt.AlignLeft)
        self.layout.addWidget(self.cursor_label, 0, 7, 1, 1)
        self.layout.addWidget(self.plot_widget, 1, 0, 1, 8)
        
        # Show or hide the menubar
        self.menubar.setVisible(self.show_menubar)
        
        # Connect signal for cursor movement
        # NOTE: Must assign to variable so it doesn't get garbage collected
        self.cursor_proxy = pg.SignalProxy(self.plot_item.scene().sigMouseMoved, rateLimit=60, 
                                           slot=self.show_cursor_position)
        
        # Hide options from the context menu that seem to be broken
        # menus = [self.plot_item.vb.menu, self.plot_item.ctrlMenu, self.plot_item.vb.scene().contextMenu[0]]
        # for menu in menus:
        #     for action in menu.actions():
        #         print(action.text())
        self.plot_item.ctrlMenu.menuAction().setVisible(False)
        
        # Show widget
        if popup:
            self.setWindowFlags(Qt.Window)
            self.show()
            self.raise_()
            self.setWindowTitle(str(name) if name is not None else "Plot")
            self.resize(*_DEFAULT_SIZE)
        
    def closeEvent(self, event) -> None:
        self.setParent(None)
        self.plot_widget.close()
        
    def leaveEvent(self, event) -> None:
        self.cursor_label.setText("")
        
    @property
    def plot_item(self) -> pg.PlotItem:
        return self.plot_widget.getPlotItem()
    
    @property
    def left(self) -> pg.AxisItem:
        return self.plot_widget.getAxis("left")
    
    @property
    def bottom(self) -> pg.AxisItem:
        return self.plot_widget.getAxis("bottom")
    
    @property
    def right(self) -> pg.AxisItem:
        return self.plot_widget.getAxis("right")
    
    @property
    def top(self) -> pg.AxisItem:
        return self.plot_widget.getAxis("top")
    
    @property
    def axes(self) -> list:
        return [getattr(self, ax) for ax in _PG_AXES]
    
    def get_curve(self, color: Union[QColor, str, tuple]) -> pg.PlotCurveItem:
        """ Get an existing plot item. """
        color = get_qcolor(color)
        return self.plot_items.get(color.name())
    
    @pyqtSlot(str)
    def add_curve(self, color: Union[QColor, str, tuple]) -> pg.PlotCurveItem:
        """ Add a curve to the plot. """
        # Raise error if curve already exists
        color = get_qcolor(color)
        if self.plot_items.get(color.name()) is not None:
            raise AttributeError(f"{color.name()} curve already exists.")
            
        # Create curve (named after the color hex) and return it
        pen = get_qpen(color, cosmetic=True)
        curve = pg.PlotCurveItem(pen=pen, name=color.name())
        self.plot_items[color.name()] = curve
        self.plot_item.addItem(curve)
        
        # Emit curve_added and connect data update signal so FFT can update
        self.curve_added.emit(color.name())
        curve.sigPlotChanged.connect(lambda c: self.data_changed.emit(color.name()))
        
        # Create action in curve_menu
        self.add_curve_menu_action(color.name())
        
        return curve
    
    @pyqtSlot(str)
    def get_or_add_curve(self, color: Union[QColor, str, tuple]) -> pg.PlotCurveItem:
        """ Get a curve or add it if it doesn't exist. """
        # Return curve if it already exists
        color = get_qcolor(color)
        if self.plot_items.get(color.name()) is not None:
            return self.get_curve(color)
        
        # Create curve
        return self.add_curve(color)
    
    @pyqtSlot(str)
    def remove_curve(self, color: Union[QColor, str, tuple]) -> None:
        """ Remove a curve from the plot. """
        # Remove the curve from the plot
        color = get_qcolor(color)
        self.plot_item.removeItem(self.get_curve(color))
        
        # Remove from storage
        self.plot_items.pop(color.name(), None)
        
        # Remove action from menu
        action = self._get_menu_action(color.name())
        action.setParent(None) if action is not None else None
        
        # Emit curve_removed so FFT can update
        self.curve_removed.emit(color.name())
    
    @pyqtSlot(object)
    def show_cursor_position(self, event: object) -> None:
        # Get position of event
        pos = event[0]
        
        # Get local cursor position and update label
        try:
            cursor = self.plot_item.vb.mapSceneToView(pos)
            x, y = cursor.x(), cursor.y()
            self.cursor_label.setText(f"{x:.2f}, {y:.2f}")
            
        # Catch exception if the plot is collapsed in splitter
        except:
            pass
        
    @pyqtSlot(str, bool)
    def curve_menu_clicked(self, color: str, visible: bool) -> None:
        self.toggle_curve(color, visible)
        
    def add_curve_menu_action(self, color: str) -> QAction:
        """ Add a QAction to the curve_menu for showing/hiding a curve. """
        # Check if curve exists
        action = self._get_menu_action(color)
        if action is not None:
            return action
        
        # Create menu item 
        # 'color' should be hex value
        action = self.curve_menu.addAction(color)
        action.setCheckable(True)
        action.setChecked(True)
        
        # Update font color
        # TODO: Fiigure out how to do this
        
        # Link signal to show/hide curve
        action.toggled.connect(lambda v: self.curve_menu_clicked(color, v))
        
        return action
    
    def get_items(self) -> list:
        return self.plot_widget.listDataItems()
    
    def toggle_curve(self, color: str, visible: bool, block_signal: bool = False) -> None:
        """ Show or hide a curve. """
        # Get the curve
        curve = self.get_curve(color)
        
        # If the curve doesn't exist, return early
        if curve is None:
            return
        
        # Toggle the curve
        curve.setVisible(visible)
        
        # Emit curve_toggled signal
        self.curve_toggled.emit(color, visible) if not block_signal else None
        
    def _get_menu_action(self, color: str) -> Union[QAction, None]:
        return next((i for i in self.curve_menu.actions() if i.text() == color), None)
    

class LinePlotWidget(PlotWidget):
    def __init__(
            self, 
            parent: PlotWidget, 
            popup: bool = False, 
            name: Optional[str] = None,
            title: Optional[str] = None,
            show_menubar: bool = True,
            ) -> None:
        
        super().__init__(parent=parent, popup=popup, name=name, title=title,
                         show_menubar=show_menubar)
        
        # Update labels
        self.bottom.setLabel("Time", units="s")
        self.left.setLabel("Intensity (counts)")
        
        # Create LinearRegionItem for adjusting FFT window
        b = pg.mkBrush((0, 0, 255, 10))
        hb = pg.mkBrush((0, 0, 255, 25))
        self.fft_window = pg.LinearRegionItem(brush=b, hoverBrush=hb, bounds=(0, None))
        self.plot_item.addItem(self.fft_window)
        
        # Create input for manually setting min/max
        self.fft_min_input = QDoubleSpinBox()
        self.fft_min_input.setValue(self.fft_bounds[0])
        self.fft_max_input = QDoubleSpinBox()
        self.fft_max_input.setValue(self.fft_bounds[1])
        for box in [self.fft_min_input, self.fft_max_input]:
            box.setMinimum(0)
            box.setMaximum(9999)
            box.setDecimals(2)
            box.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)
            
        # Create labels
        self.fft_min_label = QLabel("Min:")
        self.fft_max_label = QLabel("Max:")
        [w.setAlignment(Qt.AlignRight | Qt.AlignVCenter) for w in 
         (self.fft_min_label, self.fft_max_label)]
        
        # Create checkbox for autoscale max
        self.auto_fft_max_checkbox = QCheckBox("Autoscale max")
        self.auto_fft_max_checkbox.setChecked(True)
        
        # Add widgets
        self.layout.addWidget(self.fft_min_label, 0, 1, 1, 1)
        self.layout.addWidget(self.fft_min_input, 0, 2, 1, 1)
        self.layout.addWidget(self.fft_max_label, 0, 3, 1, 1)
        self.layout.addWidget(self.fft_max_input, 0, 4, 1, 1)
        self.layout.addWidget(self.auto_fft_max_checkbox, 0, 5, 1, 1)
        
        # Connect signal for fft_window bounds changing
        self.fft_window.sigRegionChanged.connect(self.fft_window_dragged)
        self.fft_min_input.valueChanged.connect(self.set_fft_min)
        self.fft_max_input.valueChanged.connect(self.set_fft_max)
        
    @property
    def fft_bounds(self) -> list:
        return sorted([line.getXPos() for line in self.fft_window.lines])
    
    @property
    def auto_fft_max(self) -> bool:
        return self.auto_fft_max_checkbox.isChecked()
    
    @pyqtSlot()
    def fft_window_dragged(self) -> None:
        """ Update the FFT window region inputs. """
        [b.blockSignals(True) for b in [self.fft_min_input, self.fft_max_input]]
        self.fft_min_input.setValue(self.fft_bounds[0])
        self.fft_max_input.setValue(self.fft_bounds[1])
        [b.blockSignals(False) for b in [self.fft_min_input, self.fft_max_input]]
        
    @pyqtSlot(float)
    def set_fft_min(self, x: float) -> None:
        """ Set the minimum position of the FFT window. """
        self.fft_window.blockLineSignal = True
        if not self.fft_window.moving:
            self.fft_window.setRegion((x, self.fft_bounds[1]))
        self.fft_window.blockLineSignal = False
        
    @pyqtSlot(float)
    def set_fft_max(self, x: float) -> None:
        """ Set the maximum position of the FFT window. """
        self.fft_window.blockLineSignal = True
        if not self.fft_window.moving:
            self.fft_window.setRegion((self.fft_bounds[0], x))
        self.fft_window.blockLineSignal = False
        
    def get_data(self, color: str, ignore_window: bool = False) -> tuple:
        """ Get data from one of the lines. """
        # TODO
        

class FFTPlotWidget(PlotWidget):
    """ Widget for showing FFT data from another plot. """
    def __init__(
            self, 
            parent: PlotWidget, 
            popup: bool = False, 
            name: Optional[str] = None,
            title: Optional[str] = None,
            show_menubar: bool = True,
            autofind_peaks: bool = True,
            low_freq_cutoff: Optional[float] = _MIN_FFT_PEAK_POS,
            ) -> None:
        
        super().__init__(parent=parent, popup=popup, name=name, title=title, 
                         show_menubar=show_menubar)
        self._parent = parent
        self.autofind_peaks = autofind_peaks
        self.low_freq_cutoff = low_freq_cutoff
        self.vlines = {}
        
        # Create corresponding plot items
        [self.add_curve(color) for color in self._parent.plot_items]
        
        # Connect signal so that curves are added/removed correspondingly
        self._parent.curve_added.connect(self.add_curve)
        self._parent.curve_removed.connect(self.remove_curve)
        self._parent.data_changed.connect(self.plot_fft)
        self.curve_toggled.connect(self.toggle_vlines)
        
        # Update axes
        self.bottom.setLabel("Frequency", units="Hz")
        
    @pyqtSlot(str)
    def plot_fft(self, color: str) -> None:
        # Don't plot if the curve is not visible, and hide all vertical lines
        fft_curve = self.get_curve(color)
        if not fft_curve.isVisible():
            [self.plot_item.removeItem(line) for line in self.vlines.get(color, [])]
            return
        
        # Get QColor
        color = get_qcolor(color)
        
        # Get parent & FFT curves
        parent_curve = self._parent.get_curve(color)
        
        # Get curve data
        x, y = parent_curve.getData()
        
        # Apply cutoffs
        minval, maxval = self._parent.fft_bounds
        x, y = apply_cutoffs(x=x, y=y, minval=minval, maxval=maxval)
        
        # Try to compute FFT
        # NOTE: If data isn't copied, it will mess with the original curve data
        freq, psd = calc_fft(x.copy(), y.copy())
        if freq is None or psd is None:
            return
        
        # Cutoff low frequency peak
        freq, psd = self._cutoff_low_freq(freq, psd)
        
        # Update corresponding curve data
        try:
            fft_curve.setData(freq, psd)
        except RuntimeError:
            pass
        
        # Show peak positions, if option is selected
        self.detect_and_show_peaks(freq, psd, color.name()) if self.autofind_peaks else None
        
    @pyqtSlot(str, bool)
    def toggle_vlines(self, color: str, visible: bool) -> None:
        """ Show/hide lines for a particular curve. """
        [line.setVisible(visible) for line in self.vlines.get(color, [])]
        
    def detect_and_show_peaks(self, x: list, y: list, color: Optional[str] = None) -> None:
        # Find peaks
        peak_positions = detect_peaks(x, y, _MIN_FFT_PEAK_POS)
        if peak_positions is None:
            return
        
        # Clear vlines for current color
        if color is not None:
            color = get_qcolor(color).name()
            lines = self.vlines.get(color, [])
            [self.plot_item.removeItem(line) for line in lines]
            
        # Add lines
        pen = pg.mkPen()
        pen.setStyle(Qt.DashLine)
        pen.setColor(get_qcolor(color)) if color is not None else None
        new_lines = [self.plot_item.addLine(x=x, pen=pen) for x in peak_positions]
        if color is not None:
            self.vlines[color] = new_lines
            
    def _cutoff_low_freq(self, x: list, y: list) -> tuple:
        return apply_cutoffs(x=x, y=y, minval=self.low_freq_cutoff, maxval=None)


class LineProfileWidget(PlotWidget):
    """ Widget for displaying linear profile as a 2D time series """
    def __init__(
            self, 
            parent: FFTPlotWidget, 
            popup: bool = False, 
            name: Optional[str] = None,
            title: Optional[str] = None,
            show_menubar: bool = True
            ) -> None:
        super().__init__(parent=parent, popup=popup, name=name, title=title,
                         show_menubar=show_menubar)
        
        # Update labels
        self.bottom.setLabel("Position")
        self.left.setLabel("Intensity (Counts)")


class LineScanPlotWidget(PlotWidget):
    """ Widget for displaying linear profile as a 2D time series """
    def __init__(
            self, 
            parent: FFTPlotWidget, 
            popup: bool = False, 
            name: Optional[str] = None,
            title: Optional[str] = None,
            show_menubar: bool = True
            ) -> None:
        super().__init__(parent=parent, popup=popup, name=name, title=title,
                         show_menubar=show_menubar)
        
        # Update labels
        self.bottom.setLabel("Time", units="s")
        self.left.setLabel("Position")
        
        # Image placeholders
        # TODO: Handle line scans from multiple different lines
        self._image: Optional[np.ndarray] = None
        self._pixmap_item: Optional[QGraphicsPixmapItem] = None
        
        # TESTING
        # self.set_image(np.random.rand(100, 100))
        
    @property
    def image(self) -> Union[np.ndarray, None]:
        return self._image
    
    @image.setter
    def image(self, image: np.ndarray) -> None:
        # Apply colormap
        cmapped = apply_cmap(image, DEFAULT_CMAP)
        
        # Update pixmap
        self.pixmap = ndarray_to_qpixmap(cmapped)
        
        # Store image
        self._image = image
        
    @property
    def pixmap_item(self) -> QGraphicsPixmapItem:
        return self._pixmap_item
    
    @pixmap_item.setter
    def pixmap_item(self, pixmap: Union[QPixmap, QGraphicsPixmapItem]) -> None:
        if isinstance(pixmap, QPixmap) and self._pixmap_item is not None:
            # Replace the pixmap
            self._pixmap_item.setPixmap(pixmap)
            
        elif isinstance(pixmap, QPixmap) and self._pixmap_item is None:
            self._pixmap_item = QGraphicsPixmapItem(pixmap)
            # Add the image to the plot
            self.plot_item.addItem(self._pixmap_item)
            
        elif isinstance(pixmap, QGraphicsPixmapItem) and self._pixmap_item is None:
            self._pixmap_item = pixmap
            # Add the image to the plot
            self.plot_item.addItem(self._pixmap_item)
            
        elif isinstance(pixmap, QGraphicsPixmapItem) and self._pixmap_item is not None:
            self._pixmap_item.pixmap().swap(pixmap.pixmap())
            
        else:
            raise TypeError(f"Expected QPixmap or QGraphicsPixmapItem, got {type(pixmap)}")
            
    @property
    def pixmap(self) -> Union[QPixmap, None]:
        if self.pixmap_item is not None:
            return self.pixmap_item.pixmap()
        return None
    
    @pixmap.setter
    def pixmap(self, pixmap: QPixmap) -> None:
        self.pixmap_item = pixmap
        
    def set_image(self, image: np.ndarray) -> None:
        """ Set the current image. """
        # Use property setter to update the item and pixmap
        self.image = image
        
        # Scale the bounds properly
        h, w = image.shape[:2]
        self.plot_item.setXRange(0, w*1.05, padding=0)  # default between 0.02 and 0.1 if not specified
        self.plot_item.setYRange(0, h*1.02, padding=0)
        
    
class GrowthRatePlotWidget(PlotWidget):
    """ Widget for selecting visible lines on a plot. """
    def __init__(
            self, 
            parent: FFTPlotWidget, 
            popup: bool = False, 
            name: Optional[str] = None,
            title: Optional[str] = None,
            show_menubar: bool = True
            ) -> None:
        super().__init__(parent=parent, popup=popup, name=name, title=title,
                         show_menubar=show_menubar)
        
        # Update labels
        self.bottom.setLabel("Time", units="s")
        self.left.setLabel("Growth Rate (ML/s)")


class PlotGridWidget(QWidget):
    closed = pyqtSignal()
    
    """ Widget for containing the live RHEED plots and plot transformations. """
    def __init__(self, parent = None, title: Optional[str] = None, popup: bool = True):
        super().__init__(parent)
        self._parent = parent
        
        # Create layout
        self.layout = QGridLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(4)
        self.setLayout(self.layout)
        
        # Create controls layout
        self.controls_layout = QGridLayout()
        self.controls_layout.setContentsMargins(4, 4, 4, 4)
        self.controls_layout.setSpacing(4)
        
        # Create menubar
        self.menubar = QMenuBar(self)
        self.menubar.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)
        self.curve_menu = self.menubar.addMenu(_CURVE_MENU_TITLE)
        
        # Create controls buttons
        self.start_button = QPushButton("Start")  # TODO: Add icon
        self.stop_button = QPushButton("Stop")  # TODO: Add icon
        
        # Create plots layout
        self.plots_layout = QGridLayout()
        self.plots_layout.setContentsMargins(8, 8, 8, 8)
        self.plots_layout.setSpacing(4)
        
        # Create plot widgets
        self.region_plot = LinePlotWidget(parent=self, popup=False, 
                                          title="Region Intensity", show_menubar=False)
        self.region_fft_plot = FFTPlotWidget(parent=self.region_plot, popup=False, 
                                             title="Region Intensity FFT", show_menubar=False)
        self.growth_rate_plot = GrowthRatePlotWidget(parent=self.region_fft_plot, popup=False, 
                                                     title="Growth Rate", show_menubar=False)
        self.profile_plot = LineProfileWidget(parent=self, popup=False, 
                                              title="1D Line Profile", show_menubar=False)
        # self.profile_fft_plot = FFTPlotWidget(parent=self.profile_plot, popup=False,
        #                                       title="Line Profile FFT", show_menubar=False)
        self.line_scan_plot = LineScanPlotWidget(parent=self, popup=False, 
                                                 title="2D Line Scan", show_menubar=False)
        self.plot_widgets = [
            self.region_plot,
            self.region_fft_plot,
            self.growth_rate_plot,
            self.profile_plot,
            # self.profile_fft_plot,
            self.line_scan_plot,
            ]
        [setattr(w, "curve_menu", self.curve_menu) for w in self.plot_widgets]
        
        # Create containers for plots
        color, hover = "lightGrey", "grey"
        self.main_splitter = VisibleSplitter(color, hover, orientation=Qt.Horizontal)
        self.region_plots_splitter = VisibleSplitter(color, hover, orientation=Qt.Vertical)
        self.profile_plots_splitter = VisibleSplitter(color, hover, orientation=Qt.Vertical)
        
        # Add items to main layout
        self.layout.addWidget(self.menubar, 0, 0, 1, 1)
        self.layout.addLayout(self.controls_layout, 1, 0, 1, 1)
        self.layout.addLayout(self.plots_layout, 2, 0, 1, 1)
        self.controls_layout.addWidget(self.start_button, 0, 0, 1, 1)
        self.controls_layout.addWidget(self.stop_button, 0, 1, 1, 1)
        self.controls_layout.addItem(HSpacer(), 0, 2, 1, 1)
        self.layout.addWidget(self.main_splitter, 2, 0, 1, 1)
        self.main_splitter.addWidget(self.region_plots_splitter)
        self.main_splitter.addWidget(self.profile_plots_splitter)
        [self.region_plots_splitter.addWidget(w) 
         for w in (self.region_plot, self.region_fft_plot, self.growth_rate_plot)]
        [self.profile_plots_splitter.addWidget(w)
         for w in (self.profile_plot, self.line_scan_plot)]
        
        # Connect signals
        for widget in self.plot_widgets:
            widget.curve_toggled.connect(self.toggle_all_curves)
        
        # Resize splitter
        self.main_splitter.setSizes([350, 350])
        
        # Show widget
        _DEFAULT_SIZE = (800, 600)
        if popup:
            self.setWindowFlags(Qt.Window)
            self.show()
            self.raise_()
            self.setWindowTitle(str(title) if title is not None else "Plots")
            self.resize(*_DEFAULT_SIZE)
            
    def closeEvent(self, event) -> None:
        self.closed.emit()
        super().closeEvent(event)
        
    @pyqtSlot(str, bool)
    def toggle_all_curves(self, color: str, visible: bool) -> None:
        [wid.toggle_curve(color, visible, block_signal=True) for wid in self.plot_widgets]
        print("toggled all curves")
        
    @pyqtSlot(object)
    def remove_curves(self, shape) -> None:
        """ Remove curves from plots. """
        color = shape.color.name()
        [wid.remove_curve(color) for wid in self.plot_widgets]
        

if __name__ == "__main__":
    def test():
        from frheed.utils import test_widget
        
        return test_widget(PlotGridWidget, block=True, parent=None)

    widget, app = test()
    
    


# -*- coding: utf-8 -*-
"""
Assorted image processing operations.
"""

from typing import Union, List

import numpy as np
import cmapy
import cv2
import matplotlib as mpl
from PyQt5.QtGui import QImage, QPixmap

# https://stackoverflow.com/a/1735122/10342097
def normalize(arr: np.ndarray) -> np.ndarray:
    """
    Normalize a numpy array between 0 and 255 as uint8.
    
    Parameters
    ----------
    arr : np.ndarray
        The array to normalize.
    
    %timeit results (1536 x 2048, float32):
        6.41 ms ± 248 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)
        
    """
    
    # Return if array is already uint8
    dtype = arr.dtype
    if dtype == np.uint8:
        return arr
    
    # Cast to floating point if it isn't already
    # https://stackoverflow.com/a/1168729/10342097
    if dtype.kind in ("u", "i"):
        
        # Warning: copy=False means the original input is modified!
        arr = arr.astype(np.float32, copy=True)
        
    # Normalize between 0 and 255
    arr *= 255 / arr.max()
    
    # Convert back to uint8
    return arr.astype(np.uint8, copy=True)
    
def apply_cmap(arr: np.ndarray, cmap: str) -> np.ndarray:
    """
    Apply a named colormap to an array. This function uses the cmapy library 
    to convert matplotlib colormaps to cv2 colormaps, since cv2.applyColormap is
    approximately 5x faster than using matplotlib/numpy methods 
    (tested using uint8 2048 x 1536 arrays, ~30ms vs ~6ms).

    Parameters
    ----------
    arr : np.ndarray
        The array to apply the colormap to. The array must be single-channel (not RGB).
    cmap : str
        The name of the colormap to apply (any valid matplotlib colormap).

    Returns
    -------
    np.ndarray
        The original array with the colormap applied to it.
        The resulting array will be RGB888 (RGB where each channel is uint8).
        It will have a shape of (h, w, 3) where (h, w) are the height
        and width of the input array.

    """
    return cmapy.colorize(normalize(arr), cmap, rgb_order=True)

def to_grayscale(array: np.ndarray) -> np.ndarray:
    # Get number of channels
    shape = array.shape
    channels = 1 if len(shape) == 2 else shape[-1]
    
    # Convert to grayscale if image is 3-channel
    if channels == 3:
        return cv2.cvtColor(array, cv2.COLOR_BGR2GRAY)
    
    # Convert to grayscale if RGBA
    elif channels == 4:
        return cv2.cvtColor(array, cv2.COLOR_RGBA2GRAY)
    
    # Otherwise, assume already grayscale
    return array

def ndarray_to_qimage(array: np.ndarray) -> QImage:
    """ Convert a grayscale image to a QImage. """
    # Copy the array otherwise you could get an error that QImage argument 1
    # has unexpected type 'memoryview'
    array = array.copy()
    
    # Convert to QImage
    h, w = array.shape[0:2]
    bytes_per_line = array.strides[0]  # assuminng C-contiguous array
    return QImage(array.data, w, h, bytes_per_line, QImage.Format_RGB888)

def ndarray_to_qpixmap(array: np.ndarray) -> QPixmap:
    """ Convert a numpy array to a QPixmap. """
    return QPixmap(ndarray_to_qimage(array))

def column_to_image(column: Union[np.ndarray, list]) -> np.ndarray:
    # Convert column to ndarray
    column = np.array(column)
    
    # Convert to 2D array
    return column[::-1, np.newaxis]

def extend_image(image: np.ndarray, new_col: np.ndarray) -> np.ndarray:
    """ Append a new column onto the right side of an image. """
    # Make sure the new column is the same height as the image
    # If it isn't, pad the edges with np.nan
    h, w = image.shape[:2]
    if new_col.size != h:
        print(f"Image height {h} does not match column height {new_col.size}")
        return column_to_image(new_col)
    
    return np.append(image, column_to_image(new_col), axis=1)

def get_valid_colormaps() -> List[str]:
    return mpl.pyplot.colormaps()


if __name__ == "__main__":
    
    import time
    
    from FRHEED.utils import sample_array
    
    class NormTests:
        def norm_0(arr: np.ndarray) -> np.ndarray:
            # 7.04 ms ± 235 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)
            arr = arr.astype(np.float32, copy=False)
            arr *= 255 / arr.max()
            return arr.astype(np.uint8, copy=False)
        
        def norm_1(arr: np.ndarray) -> np.ndarray:
            # 12.3 ms ± 170 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)
            return (arr * (255 / arr.max())).astype(np.uint8, copy=False)
            
        def norm_2(arr: np.ndarray) -> np.ndarray:
            # 12.3 ms ± 394 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)
            info = np.iinfo(arr.dtype)
            arr = 255 * (arr.astype(np.float32, copy=False) / info.max)
            arr = arr.astype(np.uint8, copy=False)
            return arr
    
    c = 1
    # arr_uint8 = sample_array(channels=c, dtype="uint8")
    # arr_uint16 = sample_array(channels=c, dtype="uint16")
    arr_float32 = sample_array(channels=c, dtype="float32")
    
    t0 = time.time()
    normed = normalize(arr_float32)
    print(f"Normalized array in {time.time()-t0:.5f} seconds")
    
    t0 = time.time()
    cmap = "Spectral"
    mapped = apply_cmap(normed, cmap)
    print(f"Applied colormap in {time.time()-t0:.5f} seconds")
    
    print(get_valid_colormaps())



# -*- coding: utf-8 -*-
"""
Connecting to USB cameras.
"""

import os
from typing import Union, Optional, List, Tuple
import time
import cv2
import numpy as np

from frheed.cameras import CameraError

# Suppress warning from MSMF backend bug
# https://stackoverflow.com/a/54585850/10342097
os.environ["OPENCV_VIDEOIO_PRIORITY_MSMF"] = "0"

_DEBUG = (__name__ == "__main__")

# Get non-platform-specific capture properties (0 < id < 50)
_CAP_PROPS = [prop for prop in dir(cv2) if prop.startswith("CAP_PROP")
              and getattr(cv2, prop, 1000) < 50]

# Editable camera settings to show
_GUI_SETTINGS = {
    "CAP_PROP_BRIGHTNESS":      "Brightness",
    "CAP_PROP_CONTRAST":        "Contrast",
    "CAP_PROP_SATURATION":      "Saturation",
    "CAP_PROP_GAIN":            "Gain",
    "CAP_PROP_EXPOSURE":        "Exposure",
    }

# Read-only camera information to show
_GUI_INFO = {
    "CAP_PROP_FRAME_WIDTH":     "Frame Width",
    "CAP_PROP_FRAME_HEIGHT":    "Frame Height",
    "CAP_PROP_TEMPERATURE":     "Temperature",  # TODO: figure out units
    
    }

# Backend for cameras
_DEFAULT_BACKEND = cv2.CAP_DSHOW  # cv2.CAP_DSHOW or cv2.CAP_MSMF

def list_cameras() -> List[int]:
    """
    Get list of indices of available USB cameras.

    Returns
    -------
    List[int]
        List of indices of available USB cameras.

    """
    
    # Add camera indices until list is exhausted
    cam_list = []
    
    for idx in range(100):
        cap = cv2.VideoCapture(idx, _DEFAULT_BACKEND)
        if not cap.read()[0]:
            break
        else:
            cam_list.append(idx)
        cap.release()
    
    # This seems to help fix FPS when using backend CAP_DSHOW
    cv2.destroyAllWindows()
        
    return cam_list

def get_available_cameras() -> dict:
    """ Get available cameras as a dictionary of {source: name}. """
    cams = list_cameras()
    available = {}
    
    for src in cams:
        try:
            with UsbCamera(src=src) as cam:
                if cam.initialized:
                    available[src] = str(cam)
                else:
                    print(f"USB camera {src} is not available")
        except CameraError:
            print("No USB cameras detected")
            
    return available


class UsbCamera:
    """
    A class used to encapsulate a cv2.VideoCapture camera.
    
    Attributes
    ----------
    cam : cv2.VideoCapture
    running : bool
        True if acquiring images
    camera_attributes : dictionary
        Contains all of the non-platform-specific cv2.CAP_PROP_... attributes
    camera_methods : dictionary
        Contains all of the camera methods
    
    """
    
    _cap_props = {
        "CAP_PROP_POS_MSEC":         "Current position of the video file in milliseconds.",
        "CAP_PROP_POS_FRAMES":       "0-based index of the frame to be decoded/captured next.",
        "CAP_PROP_POS_AVI_RATIO":    "Relative position of the video file: 0 - start, 1 - end.",
        "CAP_PROP_FRAME_WIDTH":      "Width of the frames in the video stream.",
        "CAP_PROP_FRAME_HEIGHT":     "Height of the frames in the video stream.",
        "CAP_PROP_FPS":              "Frame rate.",
        "CAP_PROP_FOURCC":           "4-character code of codec.",
        "CAP_PROP_FRAME_COUNT":      "Number of frames in the video file.",
        "CAP_PROP_FORMAT":           "Format of the Mat objects returned by retrieve().",
        "CAP_PROP_MODE":             "Backend-specific value indicating the current capture mode.",
        "CAP_PROP_BRIGHTNESS":       "Brightness of the image (only for cameras).",
        "CAP_PROP_CONTRAST":         "Contrast of the image (only for cameras).",
        "CAP_PROP_SATURATION":       "Saturation of the image (only for cameras).",
        "CAP_PROP_HUE":              "Hue of the image (only for cameras).",
        "CAP_PROP_GAIN":             "Gain of the image (only for cameras).",
        "CAP_PROP_EXPOSURE":         "Exposure (only for cameras).",
        "CAP_PROP_CONVERT_RGB":      "Boolean flags indicating whether images should be converted to RGB.",
        "CAP_PROP_RECTIFICATION":    "Rectification flag for stereo cameras.",
        "CAP_PROP_ISO_SPEED":        "The ISO speed of the camera.",
        "CAP_PROP_BUFFERSIZE":       "Amount of frames stored in internal buffer memory.",
        }
    
    def __init__(
            self, 
            src: Union[int, str] = 0, 
            lock: bool = False,
            backend: Optional[int] = _DEFAULT_BACKEND # cv2.CAP_DSHOW
            ):
        """
        Parameters
        ----------
        src : Union[int, str], optional
            Camera source as an index or path to video file. The default is 0.
        lock : bool, optional
            If True, attribute access is locked down; after the camera is
            initialized, attempts to set new attributes will raise an error. 
            The default is True.
        backend : Optional[int], optional
            Camera backend. The default is cv2.CAP_DSHOW.
        """
        super().__setattr__("camera_attributes", {})
        super().__setattr__("camera_methods", {})
        super().__setattr__("lock", lock)
        
        cam_list = list_cameras()
        if _DEBUG:
            print(f"Found {len(cam_list)} USB camera(s)")
        
        if not cam_list:
            raise CameraError("No USB cameras detected.")
        
        self._src_type = type(src)
        self._src = src
        
        # Initialize the camera, either by index or filepath (to video)
        if backend is not None:
            self._cam = cv2.VideoCapture(src, backend)
        else:
            self._cam = cv2.VideoCapture(src)
            
        # Get camera attributes
        for attr in _CAP_PROPS:
            self.camera_attributes[attr] = {}
            if attr in self._cap_props:
                self.camera_attributes[attr]["description"] = self._cap_props[attr]
        
        # Other attributes which may be accessed later
        self._running = True  # camera is running as soon as you connect to it
        self._frame_times = []
        self._incomplete_image_count = 0
        
    def __getattr__(self, attr: str) -> object:
        # Add this in so @property decorator works as expected
        if attr in self.__dict__:
            return self.__dict__[attr]
        
        elif attr in self.camera_attributes:
            propId = getattr(cv2, attr, None)
            if propId is None:
                raise AttributeError(f"{attr} is not a valid propId")
            return self.cam.get(propId)
        
        else:
            raise AttributeError(attr)
            
    def __setattr__(self, attr: str, val: object) -> None:
        if attr in self.camera_attributes:
            propId = getattr(cv2, attr, None)
            if propId is None:
                raise AttributeError(f"Unknown propId '{attr}'")
                
            # In order to change CAP_PROP_EXPOSURE, it has to be set to 0.25
            # first in order to enable manual exposure
            # https://github.com/opencv/opencv/issues/9738#issuecomment-346584044
            if propId in ["CAP_PROP_EXPOSURE"]:
                self.cam.set(propId, 0.25)
                
            success = self.cam.set(propId, val)
            result = "succeeded" if success else "failed"
            if _DEBUG or not success:
                print(f"Setting {attr} to {val} {result}")
            
        else:
            if attr == "__class__":
                super().__setattr__(attr, val)
            elif attr not in self.__dict__ and self.lock and self.initialized:
                raise CameraError(f"Unknown property '{attr}'")
            else:
                super().__setattr__(attr, val)
        
    def __enter__(self) -> "UsbCamera":
        self.init()
        return self
    
    def __exit__(self, type, value, traceback) -> None:
        self.close()
        
    def __del__(self) -> None:
        self.close()
        
    def __str__(self) -> str:
        return f"USB (Port {self._src})"
    
    @property
    def name(self) -> str:
        return f"USB{self._src}"
    
    @property
    def camera_type(self) -> str:
        return "USB"
    
    @property
    def cam(self) -> cv2.VideoCapture:
        return self._cam

    @property
    def initialized(self) -> bool:
        return self.cam.isOpened()

    @property
    def running(self) -> bool:
        return self._running
    
    @property
    def incomplete_image_count(self) -> int:
        return self._incomplete_image_count
    
    @property
    def real_fps(self) -> float:
        """ Get the real frames per second (Hz) """
        
        # When not enough frames have been captured
        if len(self._frame_times) <= 1:
            return 0.
        
        # When fewer than 60 frames have been captured in this acquisition
        elif len(self._frame_times) < 60:
            dt = (self._frame_times[-1] - self._frame_times[0])
            return len(self._frame_times) / max(dt, 1)
        
        # Return the average frame time of the last 60 frames
        else:
            return 60 / (self._frame_times[-1] - self._frame_times[-60])
    
    @property
    def width(self) -> int:
        return int(self.CAP_PROP_FRAME_WIDTH)
    
    @property
    def height(self) -> int:
        return int(self.CAP_PROP_FRAME_HEIGHT)
    
    @property
    def shape(self) -> Tuple[int, int]:
        return (self.width, self.height)
    
    def init(self):
        if not self.initialized:
            self.cam.open(self._src)
    
    def start(self, continuous: bool = True) -> None:
        # Initialize the camera
        if not self.initialized:
            self.init()
            
        # Begin acquisition
        self._running = True
        
    def stop(self) -> None:
        self._frame_times = []
        self._incomplete_image_count = 0
        self._running = False
    
    def close(self) -> None:
        self.stop()
        self.cam.release()
        
    def get_array(self, complete_frames_only: bool = False) -> np.ndarray:
        # Grab and retrieve the camera array
        is_complete, array = self.cam.read()
        
        # Increment incomplete image count if full image is not retrieved
        if not is_complete:
            self._incomplete_image_count += 1
            
        # Ensure complete image is returned if option is chosen
        if complete_frames_only and not is_complete:
            return self.get_array(complete_frames_only)
        
        # Store frame time for real FPS calculation
        self._frame_times.append(time.time())
        
        return array
    
    def disable_auto_exposure(self) -> None:
        self.CAP_PROP_EXPOSURE = 0.25
        
    def enable_auto_exposure(self) -> None:
        self.CAP_PROP_EXPOSURE = 0.75
        
    def edit_settings(self) -> None:
        """ 
        Launch the platform-controlled webcam settings.
        This is only available with the cv2.CAP_DSHOW backend.
        """
        self.CAP_PROP_SETTINGS = 0
        
    def get_info(self, name: str) -> dict:
        info = {"name": name}
        
        
        return info


if __name__ == "__main__":
    
    def test():
        from PIL import Image 
        
        with UsbCamera() as cam:
            while True:
                try:
                    
                    for prop in cam._cap_props:
                        print(prop, (getattr(cam, prop)))
                        
                    break
                    
                    array = cam.get_array()
                    Image.fromarray(array).show()
                    break
                    
                except KeyboardInterrupt:
                    break
        
    test()
        



# -*- coding: utf-8 -*-
"""
Styling for the PyQt5 interfaces and widgets.
"""

from dataclasses import dataclass

from PyQt5.QtGui import QPalette, QColor


@dataclass
class PaletteColor:
    role: str
    

class Palette:
    """ 
    A class to simplify easy QPalette creation
    https://doc.qt.io/qt-5/qpalette.html
    
    """
    
    


# -*- coding: utf-8 -*-

"""
Install the spinnaker .whl file located in /spinnaker/...
"""

import os
from pathlib import Path
import re

from frheed.utils import install_whl, get_logger


logger = get_logger()


def install_pyspin() -> None:
    """ Install PySpin from the appropriate .whl file """
    
    # Check if PySpin is already installed
    try:
        import PySpin
        if __name__ == "__main__":
            logger.info("PySpin is already installed")
        return 
    
    except ImportError:
        pass
    
    from frheed.utils import get_platform_bitsize
    
    # Determine bitsize of platform (32- and 64-bit Windows is supported currently)
    bitsize = get_platform_bitsize()
    
    if bitsize not in [32, 64]:
        msg = f"Unsupported platform bitsize: {bitsize}"
        logger.exception(msg)
        raise ValueError(msg)
        
    # Get proper .whl file
    whl_filepaths = []
    for root, dirs, files in os.walk(os.path.dirname(__file__)):
        for file in files:
            if re.search(fr"{bitsize}.*\.whl$", file):
                whl_filepaths.append(os.path.join(root, file))
        
    if not whl_filepaths:
        msg = f"Unable to find any .whl files for platform bitsize {bitsize}"
        logger.exception(msg)
        raise ValueError(msg)
        
    pretty_filepaths = "\n\t".join(whl_filepaths)
    logger.info(f"Found .whl files:\n{pretty_filepaths}")
    
    # Try to install each of the .whl files via subprocess
    # NOTE: Different .whl files are for different versions; try each one
    # https://stackoverflow.com/a/50255019/10342097
    for file in whl_filepaths:
        logger.info(f"Trying to install PySpin from {file}")
        exit_code = install_whl(file)
        whl_filepath = file
        if exit_code == 0:
            break

    # Verify that PySpin has been installed
    try:
        import PySpin
        logger.info(f"PySpin installed successfully from {whl_filepath}")
        
    except ImportError:
        logger.info(f"PySpin failed to install from {whl_filepath}")

if __name__ == "__main__":
    def test():
        os.chdir(str(Path(__file__).parents[3]))
        install_pyspin()
        
    test()



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages, Command
import sys
import os
from shutil import rmtree
from pathlib import Path

# Package metadata
NAME = "FRHEED"
DESCRIPTION = "An open source GUI for real-time RHEED analysis."
AUTHOR = "Elliot Young"
AUTHOR_EMAIL = "elliotyoung@frheed.com"
VERSION = "0.0.1"
URL = "https://github.com/ecyoung3/FRHEED"
PYTHON_REQUIRES = "==3.8.10"  # TODO: Find minimum working verssion

# Load requirements
here = os.path.abspath(os.path.dirname(__file__))
requirements_file = os.path.join(here, "requirements.txt")
INSTALL_REQUIRES = Path(requirements_file).read_text().split("\n")

# Load version info from __version__.py
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION
    

# https://github.com/kennethreitz/setup.py/blob/master/setup.py 
class UploadCommand(Command):
    """ Support setup.py uploading to PyPI. """
    
    description = "Build and publish the package to PyPI"
    user_options = []
    
    @staticmethod
    def status(s):
        """ Print status in bold. """
        print(f"\033[1m{s}\033[0m")
        
    def initialize_options(self):
        pass
    
    def finalize_options(self):
        pass
    
    def run(self):
        try:
            self.status("Removing previous builds...")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass
        
        self.status("Updating requirements.txt")
        from frheed.utils import gen_reqs
        gen_reqs()
        
        self.status("Building source and wheel (universal) distribution...")
        os.system(f"{sys.executable} setup.py sdist bdist_wheel --universal")
        
        self.status("Uploading the package to PyPI via Twine...")
        os.system("twine upload dist/*")
        
        self.status("Pushing git tags...")
        os.system(f"git tag v{VERSION}")
        os.system("git push --tags")
        
        sys.exit()

    
setup(
      name = NAME,
      version = about["__version__"],
      description = DESCRIPTION,
      author = AUTHOR,
      author_email = AUTHOR_EMAIL,
      url = URL,
      packages = find_packages(),
      include_package_data = True,
      # long_description = long_description,  # TODO
      long_description_content_type = "text/markdown",
      exclude_package_data = {
          "": [".gitignore"],
          },
      setup_requires = [
          "setuptools-git"
          ],
      install_requires = INSTALL_REQUIRES,
      classifiers = [
        "Programming Language :: Python :: 3.8",
        ],
      cmdclass = {
          "upload": UploadCommand
          },
      )


# -*- coding: utf-8 -*-
"""
PyQt widgets for camera streaming and settings.
https://stackoverflow.com/a/33453124/10342097
"""

import os
from typing import Union
import time

from datetime import datetime
import numpy as np
import cv2

from PyQt5.QtWidgets import (
    QFrame, 
    QGridLayout, 
    QSizePolicy, 
    QScrollArea,
    QLabel, 
    QSlider, 
    QPushButton, 
    QWidget,
    QStatusBar,
    QDoubleSpinBox,
    QCheckBox,
    QComboBox,
    QLineEdit,
    QInputDialog,
    QMessageBox,
    QApplication,
    
    )
from PyQt5.QtGui import (
    QCloseEvent, 
    QImage, 
    QPixmap,
    
    )
from PyQt5.QtCore import (
    pyqtSignal, 
    pyqtSlot, 
    QObject, 
    QThread, 
    QSize, 
    Qt,
    
    )

from frheed.widgets.common_widgets import SliderLabel, DoubleSlider, HLine
from frheed.widgets.canvas_widget import CanvasWidget
from frheed.cameras import CameraError
from frheed.cameras.flir import FlirCamera
from frheed.cameras.usb import UsbCamera
from frheed.image_processing import (
    apply_cmap, to_grayscale, ndarray_to_qpixmap, extend_image, column_to_image,
    get_valid_colormaps,
    )
from frheed.constants import DATA_DIR
from frheed.utils import load_settings, save_settings
    

MIN_ZOOM = 0.20
MAX_ZOOM = 2.0
MIN_W = 480
MIN_H = 348
MAX_W = 2560
MAX_H = 2560
DEFAULT_CMAP = "Spectral"
DEFAULT_INTERPOLATION = cv2.INTER_CUBIC


class VideoWidget(QWidget):
    """ Holds the camera frame and toolbar buttons """
    
    frame_changed = pyqtSignal()
    frame_ready = pyqtSignal(np.ndarray)
    _min_w = 480
    _min_h = 348
    _max_w = MAX_W
    
    def __init__(self, camera: Union[FlirCamera, UsbCamera], parent = None):
        super().__init__(parent)
        
        # Store colormap
        self._colormap = DEFAULT_CMAP
        
        # Store camera reference and start the camera
        self.set_camera(camera)
        
        # Frame settings
        self.setSizePolicy(QSizePolicy.MinimumExpanding, 
                           QSizePolicy.MinimumExpanding)
        self.setMinimumSize(QSize(MIN_W, MIN_H))
        self.setMouseTracking(True)
        
        # Create main layout (2 rows x 1 column)
        self.layout = QGridLayout()
        self.layout.setContentsMargins(4, 4, 4, 4)
        self.layout.setSpacing(4)
        self.setLayout(self.layout)
        
        # Create toolbar layout
        self.toolbar_layout = QGridLayout()
        self.toolbar_layout.setContentsMargins(0, 0, 0, 0)
        self.toolbar_layout.setSpacing(4)
        
        # Create capture button
        self.capture_button = QPushButton("Capture")
        self.capture_button.setSizePolicy(QSizePolicy.Maximum, 
                                          QSizePolicy.Maximum)
        
        # Create start/stop button
        self.play_button = QPushButton("Stop Camera")
        self.play_button.setSizePolicy(QSizePolicy.Maximum,
                                       QSizePolicy.Maximum)
        
        # Determine maximum zoom because huge images lag the system
        cam_w, cam_h = camera.width, camera.height
        max_zoom = max(min(MAX_ZOOM, (MAX_W / cam_w), (MAX_H / cam_h)), 1)
        
        # Create zoom slider
        self.slider = DoubleSlider(decimals=2, log=False, parent=self)
        self.slider.setFocusPolicy(Qt.NoFocus)
        self.slider.setMinimum(MIN_ZOOM)
        self.slider.setMaximum(max_zoom)
        self.slider.setValue(1.0)
        self.slider.setSingleStep(0.01)
        self.slider.setTickPosition(QSlider.TicksAbove)
        self.slider.setTickInterval(0.10)
        self.slider.setOrientation(Qt.Horizontal)
        
        # Create zoom label
        self.zoom_label = SliderLabel(self.slider, name="Zoom", precision=2)
        
        # Create button for opening settings
        self.settings_button = QPushButton()
        self.settings_button.setText("Edit Settings")
        self.settings_button.setSizePolicy(QSizePolicy.Maximum,
                                           QSizePolicy.Maximum)
        
        # Create button for opening output folder
        self.folder_button = QPushButton()
        # TODO: Finish functionality of this button
        
        # Create settings widget
        self.make_camera_settings_widget()
        
        # Create display for showing camera frame
        self.display = CameraDisplay(self)
        
        # Create scroll area for holding camera frame
        self.scroll = QScrollArea()
        self.scroll.setWidgetResizable(True)
        self.scroll.setAlignment(Qt.AlignTop | Qt.AlignLeft)
        self.scroll.setFrameShape(QFrame.NoFrame)
        self.scroll.setWidget(self.display)
        self.scroll.wheelEvent = lambda evt: self.wheelEvent(evt, self.scroll)
        
        # Create status bar
        self.status_bar = CameraStatusBar(self)
        
        # Add widgets
        self.layout.addLayout(self.toolbar_layout, 0, 0, 1, 1)
        self.toolbar_layout.addWidget(self.capture_button, 0, 0, 1, 1)
        self.toolbar_layout.addWidget(self.play_button, 0, 1, 1, 1)
        self.toolbar_layout.addWidget(self.zoom_label, 0, 2, 1, 1)
        self.toolbar_layout.addWidget(self.slider, 0, 3, 1, 1)
        self.toolbar_layout.addWidget(self.settings_button, 0, 4, 1, 1)
        self.layout.addWidget(self.scroll, 1, 0, 1, 1)
        self.layout.addWidget(self.status_bar, 2, 0, 1, 1)
        
        # Connect signals
        self.capture_button.clicked.connect(self.save_image)
        self.play_button.clicked.connect(self.start_or_stop_camera)
        self.settings_button.clicked.connect(self.edit_settings)
        self.slider.valueChanged.connect(self.display.force_resize)
        self.frame_changed.connect(self.status_bar.frame_changed)
        
        # Attributes to be assigned later
        self.frame:         Union[np.ndarray, None] = None
        self.raw_frame:     Union[np.ndarray, None] = None
        self.region_data:   dict = {}
        self.analyze_frames = True
        
        # Set up the camera streaming thread
        self.camera_worker = CameraWorker(self)
        self.camera_thread = QThread()
        self.camera_worker.moveToThread(self.camera_thread)
        self.camera_worker.frame_ready.connect(self.show_frame)
        self.camera_worker.finished.connect(self.camera_thread.quit)
        self.camera_thread.started.connect(self.camera_worker.start)
        self.camera_thread.start()
        
        # Set up plotting thread
        self.analysis_worker = AnalysisWorker(self)
        self.analysis_thread = QThread()
        self.analysis_worker.moveToThread(self.analysis_thread)
        self.analysis_worker.finished.connect(self.analysis_thread.quit)
        self.analysis_thread.started.connect(self.analysis_worker.start)
        self.analysis_thread.start()
        
        # Connect other signals
        self.frame_ready.connect(self.analysis_worker.analyze_frame)
        
        # Variables to be used in properties
        self._workers = (self.camera_worker, self.analysis_worker)
        
    def __del__(self) -> None:
        self.camera.close()
        self.setParent(None)
        
    def keyPressEvent(self, event) -> None:
        """ Propagate keyPressEvent to children """
        super().keyPressEvent(event)
        self.display.canvas.keyPressEvent(event)
        
    def wheelEvent(self, event, parent = None) -> None:
        # Zoom the camera
        # TODO: Keep frame location constant under mouse while zooming
        if event.modifiers() == Qt.ControlModifier:
            old_zoom = self.slider.value()
            step = self.slider.singleStep()
            dy = event.angleDelta().y()
            
            # Zoom out if scrolling down
            if dy < 0:
                new_zoom = max(self.slider.minimum(), old_zoom - step)
                
            # Zoom in if scrolling up
            elif dy > 0:
                new_zoom = min(self.slider.maximum(), old_zoom + step)
                
            # Set the new zoom value
            self.slider.setValue(new_zoom)
                
        # Make sure scroll area doesn't scroll while CTRL is pressed
        elif parent is not None:
            super(type(parent), parent).wheelEvent(event)
            
        # Make sure other wheelEvents function normally
        else:
            super().wheelEvent(event)
            
    def closeEvent(self, event: QCloseEvent) -> None:
        """ Stop the camera and close settings when the widget is closed """
        [worker.stop() for worker in self.workers]
        self.settings_widget.deleteLater()
        
    @pyqtSlot()
    def start_or_stop_camera(self) -> None:
        if self.camera.running:
            self.camera.stop()
            self.play_button.setText("Start Camera")
        else:
            self.camera.start(continuous=True)
            self.play_button.setText("Stop Camera")
        
    @pyqtSlot(np.ndarray)
    def show_frame(self, frame: np.ndarray) -> None:
        """ Show the next camera frame """
        # Store raw frame
        self.raw_frame = frame.copy()
        
        # Resize to display size and get dimensions
        frame = self._resize_frame(frame)
        shape = frame.shape
        h, w = shape[0:2]
        
        # Convert to grayscale
        frame = to_grayscale(frame)
            
        # Emit the frame if analysis is needed
        self.frame_ready.emit(frame) if self.analyze_frames else None
        
        # Apply colormap
        frame = apply_cmap(frame, self.colormap)
        
        # Store the processed frame
        self.frame = frame.copy()
        
        # Create QPixmap from numpy array
        qpix = ndarray_to_qpixmap(frame)
        
        # Show the QPixmap
        self.display.label.setPixmap(qpix)
        
        # Emit frame_changed signal
        self.frame_changed.emit()
        
    @pyqtSlot()
    def save_image(self) -> None:
        """ Save the currently displayed frame """
        frame = self.frame.copy()
        
        # Generate filename
        tstamp = datetime.now().strftime("%d-%b-%Y_%H%M%S")
        filename = f"{tstamp}.png"
        filepath = os.path.join(DATA_DIR, filename)
        
        # Save the image
        cv2.imwrite(filepath, frame)
        
    @pyqtSlot()
    def edit_settings(self) -> None:
        if hasattr(self.camera, "edit_settings"):
            self.camera.edit_settings()
        else:
            self.settings_widget.show()
        
    @pyqtSlot(str)
    def set_colormap(self, colormap: str) -> None:
        self.colormap = colormap
        
    @property
    def camera(self) -> Union[FlirCamera, UsbCamera]:
        return self._camera
    
    @camera.setter
    def camera(self, camera: Union[FlirCamera, UsbCamera]) -> None:
        self.set_camera(camera)
        # TODO: Fully implement this and test it
        
    @property
    def zoom(self) -> float:
        return self.slider.value()
        
    @property
    def workers(self) -> tuple:
        return self._workers
    
    @property
    def app(self) -> QApplication:
        return QApplication.instance()
    
    @property
    def colormap(self) -> str:
        return self._colormap
    
    @colormap.setter
    def colormap(self, colormap: str) -> None:
        if colormap in get_valid_colormaps():
            self._colormap = colormap
            # TODO: Update label that shows current colormap
    
    def make_camera_settings_widget(self) -> None:
        self.settings_widget = CameraSettingsWidget(self)
    
    def set_camera(self, camera: Union[FlirCamera, UsbCamera]) -> None:
        # Change the camera and start it
        self._camera = camera
        self._camera.start(continuous=True)
        
        # Update the zoom slider (if it has been created)
        if not hasattr(self, "slider"):
            return
        cam_w, cam_h = camera.width, camera.height
        max_zoom = max(min(MAX_ZOOM, (MAX_W / cam_w), (MAX_H / cam_h)), 1)
        self.slider.setMaximum(max_zoom)
        
        # Reset to 100% zoom
        self.slider.setValue(1.00)
        
        # Update camera settings widget
        self.make_camera_settings_widget()
        
    def start_analyzing_frames(self) -> None:
        self.analyze_frames = True
        
    def stop_analyzing_frames(self) -> None:
        self.analyze_frames = False
        
    def _start_workers(self) -> None:
        [worker.start() for worker in self._workers]
        
    def _stop_workers(self) -> None:
        [worker.stop() for worker in self._workers]
        
    def _resize_frame(self, frame: np.ndarray, interp: int = DEFAULT_INTERPOLATION) -> np.ndarray:
        size = self.display.sizeHint()
        w, h = size.width(), size.height()
        return cv2.resize(frame, dsize=(w, h), interpolation=interp)
    

class CameraDisplay(QWidget):
    """ Displays the camera frame itself """
    
    def __init__(self, parent: VideoWidget):
        super().__init__(parent)
        self._parent = parent
        
        # Settings
        self.setMouseTracking(True)
        self.setSizePolicy(QSizePolicy.MinimumExpanding,
                           QSizePolicy.MinimumExpanding)
        
        # Create layout
        self.layout = QGridLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.layout.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        self.setLayout(self.layout)
        
        # Create display label
        self.label = QLabel()
        self.label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        
        # Create canvas
        self.canvas = CanvasWidget(self, shape_limit=6)
        # self.canvas.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        
        # Add widgets on top of each other
        self.layout.addWidget(self.label, 0, 0, 1, 1)
        self.layout.addWidget(self.canvas, 0, 0, 1, 1)
        
        # # Stretch so widget stays in top left
        self.layout.setRowStretch(1, 1)
        self.layout.setColumnStretch(1, 1)
        
        # Bring the canvas to the front
        self.canvas.raise_()
        
    @property
    def raw_frame(self) -> Union[None, np.ndarray]:
        return self._parent.raw_frame
    
    @property
    def zoom(self) -> float:
        return self._parent.zoom
    
    def minimumSizeHint(self) -> QSize:
        """ Define the minimum dimensions of the widget """
        return QSize(MIN_W, MIN_H)
    
    def maximumSizeHint(self) -> QSize:
        return QSize(MAX_W, MAX_H)
    
    def resizeEvent(self, event) -> None:
        """ Try to force the widget to stay in the top left corner """
        super().resizeEvent(event)
        
    def paintEvent(self, event) -> None:
        super().paintEvent(event)
        self.label.resize(self.sizeHint())
        self.canvas.resize(self.sizeHint())
        # self.canvas.raise_()
        # print(self.canvas.size(), self.label.size(), self.size())
        
    def sizeHint(self) -> QSize:
        if getattr(self, "raw_frame", None) is not None:
            width = int(self.raw_frame.shape[1] * self.zoom)
            height = int(self.raw_frame.shape[0] * self.zoom)
            return QSize(width, height)
        else:
            return super().sizeHint()
        
    @pyqtSlot()
    def force_resize(self) -> None:
        """ If the FPS is low, sometimes the display can be unresponsive """
        return
        # Resize the widget
        self.resize(self.sizeHint())
        
        # Resize the image
        self.label.setPixmap(self.label.pixmap().scaled(self.sizeHint(),
                                                        Qt.KeepAspectRatio,
                                                        Qt.SmoothTransformation))
        
    def parent(self) -> Union[None, VideoWidget]:
        return self._parent or super().parent()


class CameraSettingsWidget(QWidget):
    """ A popup widget that shows camera settings """
    
    def __init__(self, parent: VideoWidget = None, popup: bool = True):
        super().__init__(parent)
        self._parent = parent
        
        # Make widget open in separate window if "popup" is True
        self.setWindowFlags(Qt.Window) if popup else None
        self.setWindowTitle("Edit Camera Settings") if popup else None
        
        # Settings
        self.setMouseTracking(True)
        
        # Create status bar for showing status / errors
        self.status_bar = QStatusBar()
        self.status_bar.setSizeGripEnabled(False)
        self.status_bar.setStyleSheet(
            self.status_bar.styleSheet() 
            + "font-style: italic; "
            + "font-size: 11px; "
            )
        self.status_bar.setContentsMargins(0, 0, 0, 0)
        
        # Create main layout
        self.layout = QGridLayout()
        self.layout.setContentsMargins(8, 8, 8, 8)
        self.layout.setSpacing(8)
        self.setLayout(self.layout)
        
        # Create label for QComboBox
        self.config_label = QLabel()
        self.config_label.setText("Configuration:")
        self.config_label.setSizePolicy(QSizePolicy.Maximum,
                                        QSizePolicy.Maximum)
        
        # Create QComboBox for selelecting configuration
        self.config_box = QComboBox()
        self.config_box.setToolTip("Select saved settings configuration.")
        
        # Create button for saving configuration
        self.save_button = QPushButton()
        self.save_button.setText("Save")
        self.save_button.setToolTip("Save current settings configuration.")
        
        # Create button for deleting configuration
        self.delete_button = QPushButton()
        self.delete_button.setText("Delete")
        self.delete_button.setToolTip("Delete the current settings configuration.")
        self.delete_button.setEnabled(False)
        
        # Create camera settings widgets (sort alphabetically)
        row = 2
        self._settings_widgets = {}
        for (title, info) in sorted(self.settings.items(), key=lambda i: i[0]):
            
            # Create widget
            info["title"] = title
            widget = self.CameraSettingWidget(info, self)
            self._settings_widgets[info["name"]] = widget
            
            # Add widget to layout
            self.layout.addWidget(widget, row, 0, 1, 4)
            row += 1
            
        # Add the widgets
        self.layout.addWidget(self.config_label, 0, 0, 1, 1)
        self.layout.addWidget(self.config_box, 0, 1, 1, 1)
        self.layout.addWidget(self.save_button, 0, 2, 1, 1)
        self.layout.addWidget(self.delete_button, 0, 3, 1, 1)
        self.layout.addWidget(HLine(), 1, 0, 1, 4)
        self.layout.addWidget(HLine(), row + 2, 0, 1, 4)
        self.layout.addWidget(self.status_bar, row + 3, 0, 1, 4)
        
        # Load the configurations
        self._saved = True
        self._previous_config = self.config_box.currentText()
        self.load_configs()
        
        # Connect signals
        self.config_box.currentTextChanged.connect(self.set_config)
        self.save_button.clicked.connect(self.save_config)
        self.delete_button.clicked.connect(self.delete_config)
            
        # Stretch the last row
        self.layout.setRowStretch(row + 1, 1)
        
        # Set fixed size
        width = int(self.sizeHint().width() * 1.5)
        height = self.sizeHint().height()
        self.setFixedSize(width, height)
        
    @pyqtSlot()
    def show(self) -> None:
        self.setVisible(True)
        
    @pyqtSlot()
    def hide(self) -> None:
        self.setVisible(False)
        
    @property
    def camera(self) -> Union[FlirCamera, UsbCamera, None]:
        return getattr(self._parent, "camera", None)
    
    @property
    def settings(self) -> dict:
        """ Only get read/write-accessible camera settings """
        settings = getattr(self.camera, "gui_settings", {})
        return {title: self.camera.get_info(name) for name, title in settings.items()}
    
    @property
    def current_config(self) -> str:
        return self.config_box.currentText()

    @property
    def saved(self) -> bool:
        return self._saved

    @saved.setter
    def saved(self, saved: bool) -> None:
        self._saved = saved
        self.save_button.setEnabled(not saved)        
        
    @pyqtSlot(str)
    def set_config(self, name: str) -> None:
        """ Set the current configuration by name """
        
        # If current configuration is not saved, prompt user to continue
        if not self._saved and self._previous_config:
            confirm = QMessageBox.question(self, "Confirm Changing Configuration",
                                           "Current configuration has not been saved.\n"
                                           "Would you like to save before switching?")
            if confirm == QMessageBox.Yes:
                self.save_configs()
        
        # Update previous config
        self._previous_config = self.current_config
        
        # Get the configuration
        config = self._configs.get(name, None)
        
        # Return if no configuration is selected 
        if config is None:
            self.delete_button.setEnabled(False)
            return
        self.delete_button.setEnabled(True)
        
        # Set settings widgets to the config values
        for setting, value in config.items():
            widget = self._settings_widgets.get(setting, None)
            if widget is not None:
                widget.set_value(value)
                
        # Mark as saved
        self.saved = True
                
        # Set the config box to the selected name
        for idx in range(self.config_box.count()):
            if self.config_box.itemText(idx) == name:
                self.config_box.setCurrentIndex(idx)
                break
    
    @pyqtSlot()
    def save_config(self) -> None:
        """ Save the current configuration """
        # Get current configuration as a dictionary
        config = self.to_dict()
        
        # Prompt user to input name
        name, ok = QInputDialog.getText(self, "Save Configuration",
                                        "Enter configuration name:", 
                                        QLineEdit.Normal, self.current_config)
        
        # Return if user cancels or rejects the dialog
        if not ok:
            return
        
        # Make sure a config name is entered
        if not name:
            QMessageBox.warning(self, "Warning", 
                                "A configuration name is required.")
            return
        
        # Make sure the entered config name is unique
        elif name in self._configs:
            action = "update" if name == self.current_config else "overwrite"
            confirm = QMessageBox.question(self, "Confirm Configuration Name",
                                           f"A configuration \"{name}\" already exists.\n"
                                           f"Would you like to {action} it?")
            if confirm != QMessageBox.Yes:
                return
        
        # Save configurations
        self._configs[name] = config
        self.save_configs()
        
        # Reload configurations
        self.load_configs()
        
        # Switch to the new configuration
        self.set_config(name)
    
    @pyqtSlot()
    def delete_config(self):
        """ Delete the currently selected configuration """
        # TODO: Prompt before deleting
        
        # Prompt before deleting
        confirm = QMessageBox.question(self, "Confirm Delete", 
                                       "Delete this configuration?")
        if confirm != QMessageBox.Yes:
            return
        
        # Remove the current configuration name
        config = self._configs.pop(self.current_config, None)
        if config is not None:
            self.save_configs()
            
        # Re-load configurations
        self.load_configs()
            
    def load_configs(self) -> None:
        """ Load configurations from config file """
        
        # Load the configuration file
        current_config = self.current_config
        cam_name = getattr(self.camera, "name", "camera")
        try:
            self._configs = load_settings(cam_name)
        except OSError:
            self._configs = {}
            
        # Clear current items
        self.config_box.clear()
            
        # Add items to the QComboBox with a blank item at the top
        self.config_box.addItems([""]+list(sorted(self._configs.keys())))
        
        # Set the current configuration back to what it was
        self.set_config(current_config)
        
    def save_configs(self) -> None:
        """ Save all configurations """
        save_settings(self._configs, getattr(self.camera, "name", "camera"))
        self.saved = True
        
    def to_dict(self) -> dict:
        """ Represent the current setting configuration as a dictionary """
        d = {}
        for setting, widget in self._settings_widgets.items():
            d[setting] = widget.get_value()
        return d
    
    class CameraSettingWidget(QWidget):
        def __init__(self, info: dict, parent: "CameraSettingsWidget"):
            super().__init__(parent)
            self._parent = parent
            
            # Store info
            self.info = info
            self.name = info["name"]
            self.title = info["title"]
            self.description = info.get("description", "")
            unit = info.get("unit", "")
            self.unit = "µs" if unit == "us" else unit  # fix microsecond formatting
            self.entries = info.get("entries", None)
            self.dtype = info.get("type", None)
            
            # Get info needed to determine appropriate widget type
            min_val = info.get("min", None)
            max_val = info.get("max", None)
            value = info.get("value", None)
            
            # Settings
            self.setMouseTracking(True)
            
            # Create layout
            self.layout = QGridLayout()
            self.layout.setContentsMargins(0, 0, 0, 0)
            self.layout.setSpacing(4)
            self.setLayout(self.layout)
            
            # Use a DoubleSlider if min & max provided
            if min_val is not None and max_val is not None:
                
                # Use log scale if difference between min & max is > 1,000
                log = (max_val / min_val) > 1e3 if min_val != 0 else False
                decimals = max(len(str(min_val).split(".")[-1]),
                               len(str(max_val).split(".")[-1]))
                self.widget = DoubleSlider(decimals, log=log, base=1.5, parent=self)
                self.widget.setMinimum(min_val)
                self.widget.setMaximum(max_val)
                self.widget.setValue(value)
                self.widget.setToolTip(f"minimum = {min_val:,g}, "
                                       f"maximum = {max_val:,g}")
                
            # Create QComboBox for options with listed entries
            elif self.entries is not None or self.dtype == "enum":
                self.widget = QComboBox()
                self.widget.addItems(self.entries)
                self.widget.setCurrentIndex(self.entries.index(value))
                
            # Create QCheckBox for options with enable/disable
            elif self.dtype == "bool":
                self.widget = QCheckBox()
                self.widget.setChecked(value)
                self.widget.setText(self.title)
                self.widget.setToolTip(self.description)
                
            # Create QLineEdit for string entries (such as name)
            elif self.dtype == "string":
                self.widget = QLineEdit()
                self.widget.returnPressed.connect(lambda: self.widget.clearFocus())
                
            else:
                return print(f"Unable to create widget for setting '{info['name']}'")
            
            # Create label
            if isinstance(self.widget, QSlider):
                self.label = SliderLabel(self.widget, name=self.title, 
                                         unit=self.unit, precision=2)
            else:
                self.label = QLabel()
                self.label.setText(f"{self.title}:")
            self.label.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
            self.label.setToolTip(self.description.replace(". ", ".\n"))
            
            # Hide the label if the widget is a QCheckBox since it has a label already
            if isinstance(self.widget, QCheckBox):
                self.label.setVisible(False)
                
            # Ignore wheel events for non-spinboxes
            if not isinstance(self.widget, QDoubleSpinBox):
                self.widget.wheelEvent = lambda event: None
            
            # Set size policy
            self.widget.setSizePolicy(QSizePolicy.MinimumExpanding,
                                      QSizePolicy.Maximum)
            
            # Disable the widget if currently inaccessible
            writable = "write" in info.get("access", "write")
            self.widget.setEnabled(writable)
            self.label.setEnabled(writable)
                
            # Set widget value
            self.set_value(value)
            
            # Connect signal AFTER setting the value so it doesn't trigger
            self.connect_signal()
            
            # Add widgets to layout
            self.layout.addWidget(self.label, 0, 0, 1, 1)
            self.layout.addWidget(self.widget, 0, 1, 1, 1)
            
        def connect_signal(self) -> None:
            if getattr(self, "_connected", False):
                return 
            
            if isinstance(self.widget, QDoubleSpinBox):
                sig = "valueChanged"
                
            elif isinstance(self.widget, DoubleSlider):
                sig = "doubleValueChanged"
            
            elif isinstance(self.widget, QComboBox):
                sig = "currentTextChanged"
                
            elif isinstance(self.widget, QCheckBox):
                sig = "stateChanged"
                
            elif isinstance(self.widget, QLineEdit):
                sig = "textChanged"
                
            else:
                return
            
            getattr(self.widget, sig).connect(self.change_setting)
            self._connected = True
            
        def set_value(self, value: Union[bool, str, float]) -> None:
            if isinstance(self.widget, (QDoubleSpinBox, DoubleSlider)):
                self.widget.setValue(value)
            
            elif isinstance(self.widget, QComboBox):
                self.widget.setCurrentIndex(self.info["entries"].index(value))
                
            elif isinstance(self.widget, QCheckBox):
                self.widget.setChecked(value)
                
            elif isinstance(self.widget, QLineEdit):
                self.widget.setText(str(value))
                
        def get_value(self) -> Union[bool, str, float]:
            """ Get the current value of the setting """
            if isinstance(self.widget, (QDoubleSpinBox, DoubleSlider)):
                return self.widget.value()
            
            elif isinstance(self.widget, QComboBox):
                return self.widget.currentText()
            
            elif isinstance(self.widget, QLineEdit):
                return self.widget.text()
                
            elif isinstance(self.widget, QCheckBox):
                return self.widget.isChecked()
                
        def change_setting(self, value: Union[bool, str, float]) -> None:
            # Mark the SettingsWidget as not saved
            self._parent.saved = False
            
            value = self.get_value()  # stateChanged emits a bit, not bool
            try:
                setattr(self._parent.camera, self.name, value)
                
                if isinstance(value, (float, int)):
                    message = f"Successfully set {self.name} to {value:,g}"
                else:
                    message = f"Successfully set {self.name} to {value}"
                    
                if self.unit:
                    message += f" {self.unit}"
                    
            except CameraError as cam_error:
                message = f"Error setting {self.name}: {cam_error}"
                
            except Exception as ex:
                message = f"Error setting {self.name}: {ex}"
                
            finally:
                self._parent.status_bar.showMessage(message)
                self._parent.status_bar.setToolTip(message)
                
            # TODO: Re-check if settings are editable after changing
            # them (e.g. AcquisitionFrameRateEnable)
    

class CameraStatusBar(QStatusBar):
    def __init__(self, parent: VideoWidget = None):
        super().__init__(parent)
        self._parent = parent
        
        # Disable size grip
        self.setSizeGripEnabled(False)
        
        # Create widget for displaying FPS
        self.fps_label = QLabel()
        
        # Add widget for displaying incomplete frame count
        self.incomplete_frames_label = QLabel()
        self.incomplete_frames_label.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
        
        # Add widget for displaying errors
        self.error_label = QLabel()
        self.error_label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        
        # Add widgets
        self.insertWidget(0, self.fps_label, 0)
        self.insertWidget(1, self.incomplete_frames_label, 1)
        self.insertWidget(2, self.error_label, 0)
        
        # Display status
        
        # # Remove border on widget
        # self.setStyleSheet(self.styleSheet() + """
        #                    QStatusBar::item {border: None;}
        #                    """
        #                    )
        
    @property
    def camera(self) -> Union[FlirCamera, UsbCamera, None]:
        return getattr(self._parent, "camera", None)
    
    @property
    def fps(self) -> float:
        return getattr(self.camera, "real_fps", 0.)
    
    @property
    def incomplete_image_count(self) -> str:
        return str(getattr(self.camera, "incomplete_image_count", ""))
    
    @property
    def error_status(self) -> str:
        return str(getattr(self.camera, "error_status", "No errors"))
    
    @pyqtSlot(str)
    def show_error(self, message: str) -> None:
        self.error_label.setText(str(message))
    
    @pyqtSlot()
    def frame_changed(self) -> None:
        self.fps_label.setText(f"{self.fps:.2f} Hz ")
        self.incomplete_frames_label.setText(
            f"Incomplete images: {self.incomplete_image_count}"
            )
        self.error_label.setText(self.error_status)


class Worker(QObject):
    """
    A base class for worker objects.
    https://stackoverflow.com/a/33453124/10342097 
    """
    
    finished = pyqtSignal()
    frame_ready = pyqtSignal(np.ndarray)
    data_ready = pyqtSignal(np.ndarray)
    exception = pyqtSignal(Exception)
    
    def __init__(self, parent: VideoWidget):
        super().__init__()
        self._parent = parent
        self._running = False
        
    def camera(self) -> Union[FlirCamera, UsbCamera, None]:
        return getattr(self._parent, "camera", None)
    
    def display(self) -> Union[CameraDisplay, None]:
        return getattr(self._parent, "display", None)
    
    def canvas(self) -> Union[CanvasWidget, None]:
        return getattr(self.display(), "canvas", None)
    
    def running(self) -> bool:
        return self._running


class CameraWorker(Worker):
    """ 
    A worker object to control frame acquisition.
    
    """
    @pyqtSlot()
    def start(self) -> None:
        self._running = True
        while self.running():
            try:
                if self.camera().running:
                    self.frame_ready.emit(
                        self.camera().get_array(complete_frames_only=True)
                        )
            except Exception as ex:
                self.exception.emit(ex)
    
    @pyqtSlot()
    def stop(self) -> None:
        self._running = False
        self.camera().close()
        self.finished.emit()


class AnalysisWorker(Worker):
    """
    A worker object to analyze region intensities.
    """
    data_ready = pyqtSignal(dict)
    
    data = {}
    start_time = None
    
    @property
    def shapes(self) -> Union[list, tuple]:
        return getattr(self.canvas(), "shapes", ())
    
    @property
    def raw_frame(self) -> Union[np.ndarray, None]:
        return getattr(self.camera(), "raw_frame", None)
    
    @pyqtSlot(np.ndarray)
    def analyze_frame(self, frame: np.ndarray) -> None:
        if not self.running():
            return
        
        # Get time of data collection relative to start
        if self.start_time is None or len(self.shapes) == 0:
            self.reset_timer()
        t = time.time() - self.start_time
        
        # Get pixel intensities under regions of interest
        for shape in self.shapes:

            # Get the mask
            mask = shape.mask

            # Make sure the mask and frame have the same shape
            if mask.shape != frame.shape:
                continue
            
            # Extract the data based on the mask region
            data = frame[mask]
            
            # Store the data
            color = shape.color_name
            if color not in self.data:
                self.data[color] = {
                    "time":     [],
                    "sum":      [],
                    "average":  [],
                    "x":        [],
                    "y":        [],
                    "image":    None,
                    "kind":     shape.kind,
                    }
                
            # Store time value
            self.data[color]["time"].append(t)
            
            # Store line profile
            if shape.kind == "line":
                # self.data[color]["x"].append(np.arange(0, data.size, 1))
                ydata = data.flatten()
                self.data[color]["y"].append(ydata)
                
                # Make sure line scan image exists
                img = self.data[color]["image"]
                if img is None:
                    # Turn single column into 2D array where 0 is in the bottom left
                    # (assuming starting array like [1, 2, 3, 4, 5, ...])
                    # https://stackoverflow.com/a/44772452/10342097
                    self.data[color]["image"] = column_to_image(ydata)
                    
                # Update line scan data
                else:
                    self.data[color]["image"] = extend_image(img, ydata)
                
            else:
                # Make sure sum is non-zero to avoid divide-by-zero
                mask_sum = mask.sum()
                if mask_sum != 0:
                    self.data[color]["average"].append(data.sum() / mask_sum)
            
        self.data_ready.emit(self.data.copy())
                
    @pyqtSlot()
    def start(self) -> None:
        self._running = True
        self.start_time = time.time()
            
    @pyqtSlot()
    def stop(self) -> None:
        self._running = False
        self.start_time = None
        self.finished.emit()
        
    @pyqtSlot()
    def reset(self) -> None:
        self.data = {}
        self.reset_timer()
        
    @pyqtSlot()
    def reset_timer(self) -> None:
        self.start_time = time.time()


if __name__ == "__main__":
    def test():
        from frheed.utils import test_widget
    
        camera = FlirCamera(lock=False)
        # camera = UsbCamera(lock=False)
        widget, app = test_widget(VideoWidget, camera=camera, 
                                  block=True)
        return widget, app

    widget, app = test()
    
    app.quit()



# -*- coding: utf-8 -*-
"""
Widgets for RHEED analysis.
"""

from typing import Union

from PyQt5.QtWidgets import (
    QWidget,
    QGridLayout,
    QSizePolicy,
    QPushButton,
    QSplitter,
    QSpacerItem,
    QMenuBar,
    QMenu,
    
    )
from PyQt5.QtCore import (
    Qt,
    pyqtSlot,
    pyqtSignal,
    
    )
# from PyQt5.QtGui import (
    
#     )

from frheed.widgets.camera_widget import VideoWidget
from frheed.cameras.flir import FlirCamera
from frheed.cameras.usb import UsbCamera
from frheed.widgets.plot_widgets import PlotGridWidget
from frheed.widgets.canvas_widget import CanvasShape, CanvasLine
from frheed.widgets.selection_widgets import CameraSelection
from frheed.widgets.common_widgets import HSpacer, VSpacer
from frheed.utils import snip_lists


class RHEEDWidget(QWidget):
    _initialized = False
    
    def __init__(self, parent: QWidget = None):
        super().__init__(parent)
        
        # Settings
        self.setSizePolicy(QSizePolicy.MinimumExpanding,
                           QSizePolicy.MinimumExpanding)
        
        # Create the layout
        self.layout = QGridLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(4)
        self.setLayout(self.layout)
        
        # Create the menu bar
        self.menubar = QMenuBar(self)
        
        # "File" menu
        # Note: &File underlines the "F" to indicate the keyboard shortcut,
        # but will not be visible unless enabled manually in Windows.
        # To enable it, go to Control Panel -> Ease of Access -> Keyboard 
        #                   -> Underline keyboard shortcuts and access keys
        self.file_menu = self.menubar.addMenu("&File")
        self.file_menu.addAction("&Change camera", self.show_cam_selection)
        
        # "View" menu
        self.view_menu = self.menubar.addMenu("&View")
        self.show_live_plots_item = self.view_menu.addAction("&Live plots")
        self.show_live_plots_item.setCheckable(True)
        self.show_live_plots_item.setChecked(True)
        self.show_live_plots_item.toggled.connect(self.show_live_plots)
        
        # "Tools" menu
        self.tools_menu = self.menubar.addMenu("&Tools")
        self.preferences_item = self.tools_menu.addAction("&Preferences")
        
        # Add menubar
        self.layout.addWidget(self.menubar, 0, 0, 1, 1)
        
        # Create camera selection widget and wait for choice
        self.setVisible(False)
        self.cam_selection = CameraSelection()
        self.cam_selection.camera_selected.connect(self._init_ui)
        self.cam_selection.raise_()
        
    @pyqtSlot()
    def _init_ui(self) -> None:
        """ Finish UI setup after selecting a camera. """
        # Show the widget
        self.setVisible(True)
        
        # Create the camera widget
        camera = self.cam_selection._cam
        self.camera_widget = VideoWidget(camera, parent=self)
        self.camera_widget.setSizePolicy(QSizePolicy.MinimumExpanding,
                                         QSizePolicy.MinimumExpanding)
        
        # Create the plot widgets
        # self.region_plot = PlotWidget(parent=self, popup=True, name="Regions (Live)")
        # self.profile_plot = PlotWidget(parent=self, popup=True, name="Line Profiles (Live)")
        self.plot_grid = PlotGridWidget(parent=self, title="Live Plots", popup=True)
        self.region_plot = self.plot_grid.region_plot
        self.profile_plot = self.plot_grid.profile_plot
        self.line_scan_plot = self.plot_grid.line_scan_plot
        
        # Add widgets to layout
        self.layout.addWidget(self.camera_widget, 1, 0, 1, 1)
        self.layout.setRowStretch(1, 1)
        self.layout.setColumnStretch(0, 1)
        
        # Connect signals
        self.camera_widget.analysis_worker.data_ready.connect(self.plot_data)
        self.camera_widget.display.canvas.shape_deleted.connect(self.remove_line)
        self.plot_grid.closed.connect(self.live_plots_closed)
        self.camera_widget.display.canvas.shape_deleted.connect(self.plot_grid.remove_curves)
        
        # Reconnect camera_selected signal
        self.cam_selection.camera_selected.disconnect()
        self.cam_selection.camera_selected.connect(self.change_camera)
        
        # Mark as initialized
        self._initialized = True
        
    def resizeEvent(self, event) -> None:
        super().resizeEvent(event)
        
    def closeEvent(self, event) -> None:
        if self._initialized:
            [wid.setParent(None) for wid in 
             [self.region_plot, self.profile_plot, self, self.plot_grid]]
            self.camera_widget.closeEvent(event)
        self.cam_selection.close()
        
    @pyqtSlot(dict)
    def plot_data(self, data: dict) -> None:
        """ Plot data from the camera """
        # Get data for each color in the data dictionary
        for color, color_data in data.items():
            # Add region data to the region plot
            if color_data["kind"] in ["rectangle", "ellipse"]:
                curve = self.region_plot.get_or_add_curve(color)
                # Catch RuntimeError if widget has been closed
                try:
                    curve.setData(*snip_lists(color_data["time"], color_data["average"]))
                except RuntimeError:
                    pass
                
            # Add line profile data to the profile plot and update line scan
            elif color_data["kind"] == "line":
                curve = self.profile_plot.get_or_add_curve(color)
                try:
                    curve.setData(color_data["y"][-1])
                except RuntimeError:
                    pass
                
                # Update 2D line scan image
                self.line_scan_plot.set_image(color_data["image"])
                
            # Update region window
            if self.region_plot.auto_fft_max:
                self.region_plot.set_fft_max(color_data["time"][-1])
                
    @pyqtSlot(object)
    def remove_line(self, shape: Union["CanvasShape", "CanvasLine"]) -> None:
        """ Remove a line from the plot it is part of """
        # Get the plot widget
        plot = self.profile_plot if shape.kind == "line" else self.region_plot
        
        # Remove the line
        plot.plot_widget.removeItem(plot.plot_items.pop(shape.color_name))
        self.camera_widget.analysis_worker.data.pop(shape.color_name)
        
    @pyqtSlot()
    def show_cam_selection(self) -> None:
        """ Show the camera selection window. """
        self.cam_selection.show()
        self.cam_selection.raise_()
        
    @pyqtSlot()
    def change_camera(self) -> None:
        """ Change the active camera. """
        self.camera_widget.set_camera(self.cam_selection._cam)
        
    @pyqtSlot()
    def live_plots_closed(self) -> None:
        self.show_live_plots_item.setChecked(False)
        
    @pyqtSlot(bool)
    def show_live_plots(self, visible: bool) -> None:
        self.plot_grid.setVisible(visible)
        

if __name__ == "__main__":
    def test():
        from frheed.utils import test_widget
        
        return test_widget(RHEEDWidget, block=True)
        
    widget, app = test()
